# standard
import json

# third party
import pytest

# custom
from sunwaee.modules.gen.agent import run
from sunwaee.modules.gen.engine.base import BaseEngine
from sunwaee.modules.gen.engine.types import (
    Message,
    Response,
    Role,
    StopReason,
    Tool,
    ToolCall,
    Usage,
)

### MOCK


class MockEngine(BaseEngine):
    """Returns pre-configured Response objects in sequence."""

    def __init__(self, responses: list[Response]):
        self._responses = list(responses)
        self.calls: list[tuple[list[Message], list[Tool] | None]] = []

    async def chat(
        self, messages: list[Message], tools: list[Tool] | None = None
    ) -> Response:
        self.calls.append((list(messages), tools))
        return self._responses.pop(0)

    def stream(self, messages, tools=None):
        async def _empty():
            return
            yield  # pragma: no cover

        return _empty()


def _end(
    content: str, reasoning: str | None = None, signature: str | None = None
) -> Response:
    return Response(
        content=content,
        tool_calls=None,
        usage=Usage(input_tokens=5, output_tokens=10, total_tokens=15),
        stop_reason=StopReason.END_TURN,
        reasoning_content=reasoning,
        reasoning_signature=signature,
    )


def _tool_call(name: str, arguments: dict, call_id: str = "call-1") -> Response:
    return Response(
        content=None,
        tool_calls=[ToolCall(id=call_id, name=name, arguments=arguments)],
        usage=Usage(input_tokens=5, output_tokens=10, total_tokens=15),
        stop_reason=StopReason.TOOL_USE,
    )


def _max_tokens(content: str) -> Response:
    return Response(
        content=content,
        tool_calls=None,
        usage=Usage(input_tokens=5, output_tokens=10, total_tokens=15),
        stop_reason=StopReason.MAX_TOKENS,
    )


### DUMMY


def echo_tool(message: str = "pong") -> str:
    return json.dumps({"ok": True, "data": message})


ECHO = Tool(name="echo", description="Echo a message.", parameters={}, fn=echo_tool)


### TESTS


@pytest.mark.asyncio
async def test_simple_response_no_tools():
    engine = MockEngine([_end("Hello, user!")])
    messages = [Message(role=Role.USER, content="Hi")]

    text, new_msgs = await run(messages, [], engine)

    assert text == "Hello, user!"
    assert len(engine.calls) == 1
    # new_msgs = [assistant message]
    assert len(new_msgs) == 1
    assert new_msgs[0].role == Role.ASSISTANT
    assert new_msgs[0].content == "Hello, user!"


@pytest.mark.asyncio
async def test_reasoning_stored_in_message():
    engine = MockEngine(
        [_end("Answer", reasoning="my chain of thought", signature="sig-abc")]
    )
    text, new_msgs = await run([Message(role=Role.USER, content="Q")], [], engine)

    assistant_msg = new_msgs[0]
    assert assistant_msg.reasoning_content == "my chain of thought"
    assert assistant_msg.reasoning_signature == "sig-abc"


@pytest.mark.asyncio
async def test_single_tool_call():
    engine = MockEngine(
        [
            _tool_call("echo", {"message": "hello"}),
            _end("Done!"),
        ]
    )
    text, new_msgs = await run(
        [Message(role=Role.USER, content="Call echo")], [ECHO], engine
    )

    assert text == "Done!"
    assert len(engine.calls) == 2

    roles = [m.role.value for m in new_msgs]
    assert roles == ["assistant", "tool", "assistant"]

    tool_result = new_msgs[1]
    assert tool_result.role == Role.TOOL
    assert tool_result.tool_call_id == "call-1"
    assert tool_result.content
    result_data = json.loads(tool_result.content)
    assert result_data["data"] == "hello"


@pytest.mark.asyncio
async def test_multiple_concurrent_tool_calls():
    second_tool = Tool(
        name="second",
        description="Second tool.",
        parameters={},
        fn=lambda: json.dumps({"ok": True, "data": "second result"}),
    )
    engine = MockEngine(
        [
            Response(
                content=None,
                tool_calls=[
                    ToolCall(id="c1", name="echo", arguments={"message": "first"}),
                    ToolCall(id="c2", name="second", arguments={}),
                ],
                usage=Usage(input_tokens=5, output_tokens=10, total_tokens=15),
                stop_reason=StopReason.TOOL_USE,
            ),
            _end("All done"),
        ]
    )

    text, new_msgs = await run(
        [Message(role=Role.USER, content="Call both")],
        [ECHO, second_tool],
        engine,
    )

    assert text == "All done"
    tool_results = [m for m in new_msgs if m.role.value == "tool"]
    assert len(tool_results) == 2
    ids = {m.tool_call_id for m in tool_results}
    assert ids == {"c1", "c2"}


@pytest.mark.asyncio
async def test_unknown_tool_returns_error():
    engine = MockEngine(
        [
            _tool_call("nonexistent", {}),
            _end("Sorry"),
        ]
    )
    _, new_msgs = await run([Message(role=Role.USER, content="Hi")], [ECHO], engine)

    tool_result = next(m for m in new_msgs if m.role.value == "tool")
    assert tool_result.content
    data = json.loads(tool_result.content)
    assert data["ok"] is False
    assert "nonexistent" in data["error"]


@pytest.mark.asyncio
async def test_max_iterations_returns_last_content():
    # NOTE only tool_use responses — never returns end_turn within max_iterations
    engine = MockEngine([_tool_call("echo", {"message": "x"})] * 3)
    # patch the echo tool to avoid real execution errors being the issue
    text, new_msgs = await run(
        [Message(role=Role.USER, content="Hi")],
        [ECHO],
        engine,
        max_iterations=2,
    )
    # should return the warning message since no end_turn happened
    assert "Max iterations" in text or isinstance(text, str)


@pytest.mark.asyncio
async def test_tool_exception_returns_error_message():
    def broken_tool() -> str:
        raise ValueError("Something went wrong")

    bad_tool = Tool(name="broken", description="Breaks.", parameters={}, fn=broken_tool)
    engine = MockEngine(
        [
            _tool_call("broken", {}),
            _end("Recovered"),
        ]
    )

    _, new_msgs = await run([Message(role=Role.USER, content="Hi")], [bad_tool], engine)
    tool_result = next(m for m in new_msgs if m.role.value == "tool")
    assert tool_result.content
    data = json.loads(tool_result.content)
    assert data["ok"] is False
    assert "Something went wrong" in data["error"]


@pytest.mark.asyncio
async def test_async_tool_is_awaited():
    async def async_echo(message: str = "async") -> str:
        return json.dumps({"ok": True, "data": f"async:{message}"})

    async_tool = Tool(
        name="async_echo", description="Async echo.", parameters={}, fn=async_echo
    )
    engine = MockEngine(
        [
            _tool_call("async_echo", {"message": "hi"}),
            _end("Done"),
        ]
    )

    _, new_msgs = await run(
        [Message(role=Role.USER, content="Go")], [async_tool], engine
    )
    tool_result = next(m for m in new_msgs if m.role.value == "tool")
    assert tool_result.content
    data = json.loads(tool_result.content)
    assert data["data"] == "async:hi"


@pytest.mark.asyncio
async def test_toolcall_results_populated():
    engine = MockEngine(
        [
            _tool_call("echo", {"message": "hello"}),
            _end("Done"),
        ]
    )
    _, new_msgs = await run([Message(role=Role.USER, content="Go")], [ECHO], engine)

    assistant_msg = new_msgs[0]
    assert assistant_msg.tool_calls is not None
    tc = assistant_msg.tool_calls[0]
    assert tc.results == [{"ok": True, "data": "hello"}]
    assert tc.error is None
    assert tc.duration >= 0.0


@pytest.mark.asyncio
async def test_toolcall_error_populated_on_unknown_tool():
    engine = MockEngine([_tool_call("ghost", {}), _end("Ok")])
    _, new_msgs = await run([Message(role=Role.USER, content="Hi")], [ECHO], engine)

    assert new_msgs[0].tool_calls is not None
    tc = new_msgs[0].tool_calls[0]
    assert tc.error is not None
    assert "ghost" in tc.error
    assert tc.results == []
    assert tc.duration == 0.0


@pytest.mark.asyncio
async def test_toolcall_error_populated_on_exception():
    def broken() -> str:
        raise RuntimeError("boom")

    bad = Tool(name="bad", description=".", parameters={}, fn=broken)
    engine = MockEngine([_tool_call("bad", {}), _end("Ok")])
    _, new_msgs = await run([Message(role=Role.USER, content="Hi")], [bad], engine)

    assert new_msgs[0].tool_calls is not None
    tc = new_msgs[0].tool_calls[0]
    assert tc.error == "boom"
    assert tc.duration >= 0.0


@pytest.mark.asyncio
async def test_history_not_mutated():
    """Input messages list must not be modified by the agent."""
    engine = MockEngine([_end("Hello")])
    original = [Message(role=Role.USER, content="Hi")]
    original_len = len(original)

    await run(original, [], engine)
    assert len(original) == original_len


@pytest.mark.asyncio
async def test_new_messages_excludes_input():
    """new_messages should only contain messages added during THIS run."""
    engine = MockEngine([_end("Response")])
    history = [
        Message(role=Role.USER, content="Old message"),
        Message(role=Role.ASSISTANT, content="Old response"),
        Message(role=Role.USER, content="New question"),
    ]

    _, new_msgs = await run(history, [], engine)
    # only the assistant response from this run
    assert len(new_msgs) == 1
    assert new_msgs[0].content == "Response"



# standard
import json

# third party
# custom
from sunwaee.modules.gen.tools import TOOLS
from sunwaee.core.tools import ok, err, tool


def test_tools_contains_all_modules():
    names = {t.name for t in TOOLS}
    assert "workspaces" in names
    assert "tasks" in names
    assert "notes" in names


def test_tools_order():
    names = [t.name for t in TOOLS]
    assert names.index("workspaces") < names.index("tasks")
    assert names.index("tasks") < names.index("notes")


def test_all_tools_have_callable():
    for t in TOOLS:
        assert t.fn is not None, f"{t.name} has no callable"


def test_ok_returns_json_string():
    result = ok({"key": "value"})
    parsed = json.loads(result)
    assert parsed == {"ok": True, "data": {"key": "value"}}


def test_ok_with_list():
    result = ok([1, 2, 3])
    parsed = json.loads(result)
    assert parsed["ok"] is True
    assert parsed["data"] == [1, 2, 3]


def test_ok_with_none():
    result = ok(None)
    parsed = json.loads(result)
    assert parsed["data"] is None


def test_err_returns_json_string():
    result = err("Something went wrong")
    parsed = json.loads(result)
    assert parsed == {"ok": False, "error": "Something went wrong"}


def test_tool_unknown_type_falls_back_to_string():
    """An annotation that isn't a recognized type maps to {"type": "string"}."""

    class CustomType:
        pass

    @tool("Fallback type test.")
    def my_tool(x: CustomType) -> str:
        return ""

    assert my_tool._tool.parameters["properties"]["x"] == {"type": "string"}



# standard
# third party
import typer
from rich import box
from rich.panel import Panel

# custom
from sunwaee.config import get_default_workspace, get_workspaces_base_dir
from sunwaee.core.audit import append_log
from sunwaee.core.fs import read_md
from sunwaee.core.logger import get_logger
from sunwaee.core.output import console, error, success
from sunwaee.modules.workspaces.model import Workspace

_log = get_logger("modules.workspaces")


def read(
    name: str = typer.Argument(..., help="Workspace name"),
) -> None:
    """Read a workspace."""
    base = get_workspaces_base_dir()
    workspace_dir = base / name
    if not workspace_dir.exists():
        _log.warning("Workspace not found: %r", name)
        error(f"Workspace not found: {name!r}", "NOT_FOUND")

    meta_path = workspace_dir / "metadata.md"
    if meta_path.exists():
        meta, body = read_md(meta_path)
        ws = Workspace.from_dict(name, meta, body)
    else:
        ws = Workspace(name=name)

    _log.debug("Reading workspace %r (id=%s)", ws.name, ws.id)
    append_log(name, "workspaces", "read", item_id=ws.id, item_title=name)

    def human_output() -> None:  # pragma: no cover
        from rich.markdown import Markdown

        is_default = get_default_workspace() == name
        meta_line = "\n".join([
            f"[dim]id:[/dim] {ws.id[:8]}  "
            f"[dim]default:[/dim] {'yes' if is_default else 'no'}  "
            f"[dim]created:[/dim] {ws.created_at[:10]}  "
            f"[dim]updated:[/dim] {ws.updated_at[:10]}",
            f"[dim]path:[/dim] {workspace_dir}",
        ])
        console.print(
            Panel(
                meta_line,
                title=f"[bold]{ws.name}[/bold]",
                box=box.ROUNDED,
                expand=False,
            )
        )
        if ws.body:
            console.print(Markdown(ws.body))

    data = ws.to_dict()
    data["default"] = get_default_workspace() == name
    data["path"] = str(workspace_dir)
    success(data, human_fn=human_output)



# standard
import uuid
from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import Optional

# third party
# custom


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _uuid() -> str:
    return str(uuid.uuid4())


class TaskStatus(str, Enum):
    TODO = "todo"
    IN_PROGRESS = "in_progress"
    DONE = "done"


class TaskPriority(str, Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


@dataclass
class Task:
    title: str
    body: str = ""
    status: TaskStatus = TaskStatus.TODO
    priority: TaskPriority = TaskPriority.MEDIUM
    due_date: Optional[str] = None
    parent_id: Optional[str] = None
    completed_at: Optional[str] = None
    tags: list[str] = field(default_factory=list)
    id: str = field(default_factory=_uuid)
    created_at: str = field(default_factory=_now)
    updated_at: str = field(default_factory=_now)

    def to_dict(self) -> dict:
        d = asdict(self)
        d["status"] = self.status.value
        d["priority"] = self.priority.value
        return d

    def touch(self) -> None:
        self.updated_at = _now()

    def to_frontmatter_meta(self) -> dict:
        return {
            "id": self.id,
            "title": self.title,
            "status": self.status.value,
            "priority": self.priority.value,
            "due_date": self.due_date,
            "parent_id": self.parent_id,
            "completed_at": self.completed_at,
            "tags": self.tags,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }

    @classmethod
    def from_dict(cls, meta: dict, body: str) -> "Task":
        return cls(
            id=meta.get("id", _uuid()),
            title=meta.get("title", ""),
            status=TaskStatus(meta.get("status", "todo")),
            priority=TaskPriority(meta.get("priority", "medium")),
            due_date=meta.get("due_date"),
            parent_id=meta.get("parent_id"),
            completed_at=meta.get("completed_at"),
            tags=meta.get("tags", []),
            created_at=meta.get("created_at", _now()),
            updated_at=meta.get("updated_at", _now()),
            body=body,
        )



from __future__ import annotations

# standard
from pathlib import Path
from typing import Annotated, Literal

# third party
# custom
from sunwaee.config import get_default_workspace, get_module_dir, get_workspace_dir
from sunwaee.core.fs import find_md, list_md, make_slug, read_md, search_md, write_md
from sunwaee.core.git import ensure_git_repo, ws_commit
from sunwaee.core.tools import err, ok, tool
from sunwaee.modules.notes.model import Note


def _ws(workspace: str | None) -> str:
    return workspace or get_default_workspace()


def _dir(workspace: str) -> Path:
    return get_module_dir(workspace, "notes")


def _ws_dir(workspace: str) -> Path:
    return get_workspace_dir(workspace)


@tool(
    "Manage notes. Actions: "
    "'create' (requires title), "
    "'read' (requires query), "
    "'update' (requires query, updates only provided fields), "
    "'delete' (requires query), "
    "'list' (filter by tags/query)."
)
def notes(
    action: Annotated[
        Literal["create", "read", "update", "delete", "list"], "Action to perform"
    ],
    query: Annotated[
        str | None, "Note id, slug, or title — required for read/update/delete"
    ] = None,
    workspace: Annotated[str | None, "Workspace name (uses default if omitted)"] = None,
    title: Annotated[str | None, "Note title — required for create"] = None,
    body: Annotated[str | None, "Note content in Markdown"] = None,
    tags: Annotated[
        list[str] | None, "List of tags (filter for list, replace for update)"
    ] = None,
) -> str:
    ws = _ws(workspace)
    d = _dir(ws)
    ws_d = _ws_dir(ws)
    ensure_git_repo(ws_d)

    if action == "create":
        if not title:
            return err("'title' is required for action='create'")
        note = Note(title=title, body=body or "", tags=tags or [])
        path = d / f"{make_slug(title, note.id)}.md"
        write_md(path, note.to_frontmatter_meta(), note.body)
        diff = ws_commit(ws_d, f"notes: create '{note.title}'")
        return ok(note.to_dict(), git_diff=diff)

    if action == "list":
        paths = search_md(query, d) if query else list_md(d)
        result = []
        for p in paths:
            meta, b = read_md(p)
            note = Note.from_dict(meta, b)
            if tags and not any(t in note.tags for t in tags):
                continue
            result.append(note.to_dict())
        return ok(result)

    if not query:
        return err(f"'query' is required for action={action!r}")

    if action == "read":
        path = find_md(query, d)
        if not path:
            return err(f"Note not found: {query!r}")
        meta, b = read_md(path)
        return ok(Note.from_dict(meta, b).to_dict())

    if action == "update":
        path = find_md(query, d)
        if not path:
            return err(f"Note not found: {query!r}")
        meta, existing_body = read_md(path)
        note = Note.from_dict(meta, existing_body)
        if title is not None:  # pragma: no cover
            note.title = title
        if body is not None:
            note.body = body
        if tags is not None:  # pragma: no cover
            note.tags = tags
        note.touch()
        write_md(path, note.to_frontmatter_meta(), note.body)
        diff = ws_commit(ws_d, f"notes: update '{note.title}'")
        return ok(note.to_dict(), git_diff=diff)

    if action == "delete":
        path = find_md(query, d)
        if not path:
            return err(f"Note not found: {query!r}")
        title_for_commit = path.stem
        path.unlink()
        diff = ws_commit(ws_d, f"notes: delete '{title_for_commit}'")
        return ok({"deleted": query}, git_diff=diff)

    return err(f"Unknown action: {action!r}")  # pragma: no cover


TOOLS = [notes._tool]



# standard
import shutil
from typing import Optional

# third party
import typer

# custom
from sunwaee.config import get_default_workspace, get_project_dir
from sunwaee.core.audit import append_log
from sunwaee.core.logger import get_logger
from sunwaee.core.output import console, error, success

_log = get_logger("modules.projects")


def delete(
    name: str = typer.Argument(..., help="Project name"),
    workspace: Optional[str] = typer.Option(None, "--workspace", "-w"),
    confirm: bool = typer.Option(False, "--confirm", help="Confirm deletion"),
) -> None:
    """Delete a project and all its data."""
    if not confirm:
        error("Pass --confirm to delete a project", "CONFIRMATION_REQUIRED")

    ws = workspace or get_default_workspace()
    project_dir = get_project_dir(ws, name)

    if not project_dir.exists():
        error(f"Project not found: {name!r}", "NOT_FOUND")

    append_log(ws, "projects", "delete", item_title=name)
    shutil.rmtree(project_dir)
    _log.info("Deleted project %r from %s", name, project_dir)

    success(
        {"name": name, "workspace": ws},
        human_fn=lambda: console.print(
            f"[red]Deleted[/red] project '[bold]{name}[/bold]' and all its data"
        ),
    )



# standard
import json

# third party
# custom
from sunwaee.modules.projects.tools import projects, TOOLS as PROJECT_TOOLS

WS = "personal"


def test_project_tool_list_empty(ws):
    data = json.loads(projects(action="list", workspace=WS))
    assert data["ok"] is True
    assert data["data"] == []


def test_project_tool_list(ws):
    projects(action="create", name="alpha", workspace=WS)
    data = json.loads(projects(action="list", workspace=WS))
    assert data["ok"] is True
    assert any(p["name"] == "alpha" for p in data["data"])


def test_project_tool_list_has_metadata(ws):
    projects(action="create", name="alpha", workspace=WS)
    data = json.loads(projects(action="list", workspace=WS))
    entry = data["data"][0]
    assert "id" in entry
    assert "body" in entry
    assert "workspace" in entry
    assert "default" in entry


def test_project_tool_create(ws, tmp_path):
    data = json.loads(projects(action="create", name="myproject", workspace=WS))
    assert data["ok"] is True
    assert data["data"]["name"] == "myproject"
    assert data["data"]["workspace"] == WS
    assert "id" in data["data"]
    assert (tmp_path / "workspaces" / WS / "projects" / "myproject").is_dir()


def test_project_tool_create_with_body(ws):
    data = json.loads(projects(action="create", name="myproject", workspace=WS, body="Details"))
    assert data["ok"] is True
    assert data["data"]["body"] == "Details"


def test_project_tool_create_requires_name(ws):
    data = json.loads(projects(action="create", workspace=WS))
    assert data["ok"] is False


def test_project_tool_create_duplicate_errors(ws):
    projects(action="create", name="myproject", workspace=WS)
    data = json.loads(projects(action="create", name="myproject", workspace=WS))
    assert data["ok"] is False


def test_project_tool_read(ws):
    projects(action="create", name="myproject", workspace=WS, body="My project")
    data = json.loads(projects(action="read", name="myproject", workspace=WS))
    assert data["ok"] is True
    assert data["data"]["name"] == "myproject"
    assert data["data"]["body"] == "My project"
    assert data["data"]["workspace"] == WS
    assert "id" in data["data"]
    assert "default" in data["data"]


def test_project_tool_read_requires_name(ws):
    data = json.loads(projects(action="read", workspace=WS))
    assert data["ok"] is False


def test_project_tool_read_not_found_errors(ws):
    data = json.loads(projects(action="read", name="ghost", workspace=WS))
    assert data["ok"] is False


def test_project_tool_update_body(ws):
    projects(action="create", name="myproject", workspace=WS)
    data = json.loads(projects(action="update", name="myproject", workspace=WS, body="Updated"))
    assert data["ok"] is True
    assert data["data"]["body"] == "Updated"


def test_project_tool_update_set_default(ws):
    projects(action="create", name="alpha", workspace=WS)
    projects(action="create", name="beta", workspace=WS)
    data = json.loads(projects(action="update", name="alpha", workspace=WS, set_default=True))
    assert data["ok"] is True
    assert data["data"]["default"] is True


def test_project_tool_update_requires_name(ws):
    data = json.loads(projects(action="update", workspace=WS))
    assert data["ok"] is False


def test_project_tool_update_no_changes_errors(ws):
    projects(action="create", name="myproject", workspace=WS)
    data = json.loads(projects(action="update", name="myproject", workspace=WS))
    assert data["ok"] is False


def test_project_tool_update_not_found_errors(ws):
    data = json.loads(projects(action="update", name="ghost", workspace=WS, body="x"))
    assert data["ok"] is False


def test_project_tools_registry():
    assert len(PROJECT_TOOLS) == 1
    assert PROJECT_TOOLS[0].name == "projects"
    assert PROJECT_TOOLS[0].fn is not None



# third party
import typer

# custom
from .create import create
from .delete import delete
from .list import list_tasks
from .read import read
from .update import update

app = typer.Typer(name="tasks", help="Manage tasks.", no_args_is_help=True)
app.command("create")(create)
app.command("list")(list_tasks)
app.command("read")(read)
app.command("update")(update)
app.command("delete")(delete)



# standard
# third party
# custom
from ...conftest import ok, err
from sunwaee.cli import app

WS = "personal"
_BASE = ["--workspace", WS]


def _create(runner, title, **kwargs):
    args = ["tasks", "create", "--title", title, *_BASE]
    for k, v in kwargs.items():
        args += [f"--{k.replace('_', '-')}", str(v)]
    return runner.invoke(app, args)


def test_read_by_title(runner, ws):
    _create(runner, "Read me", body="hello")
    data = ok(runner.invoke(app, ["tasks", "read", "Read me", *_BASE]))
    assert data["title"] == "Read me"
    assert data["body"] == "hello"
    assert "path" in data


def test_read_not_found(runner, ws):
    e = err(runner.invoke(app, ["tasks", "read", "ghost", *_BASE]))
    assert e["code"] == "NOT_FOUND"


def test_read_includes_workspace(runner, ws):
    _create(runner, "WS check")
    data = ok(runner.invoke(app, ["tasks", "read", "WS check", *_BASE]))
    assert data["workspace"] == WS



# standard
from pathlib import Path
from typing import Optional

# third party
# custom
from sunwaee.config import get_all_workspaces, get_default_workspace, get_module_dir
from sunwaee.core.fs import find_md, list_md, make_slug, read_md, write_md
from sunwaee.modules.notes.model import Note


def _resolve_dir(workspace: Optional[str]) -> tuple[str, Path]:
    workspace = workspace or get_default_workspace()
    return workspace, get_module_dir(workspace, "notes")


def _resolve_dirs(workspace: Optional[str]) -> list[tuple[str, Path]]:
    if workspace:
        return [_resolve_dir(workspace)]
    return [_resolve_dir(ws) for ws in get_all_workspaces()]


def _find_in_all(
    query: str, workspace: Optional[str]
) -> Optional[tuple[str, Path, Path]]:
    """Search for a note across workspaces. Returns (ws_name, directory, path) or None."""
    for ws, directory in _resolve_dirs(workspace):
        path = find_md(query, directory)
        if path is not None:
            return ws, directory, path
    return None


def _load(path: Path) -> Note:
    meta, body = read_md(path)
    return Note.from_dict(meta, body)


def _save(note: Note, directory: Path, path: Optional[Path] = None) -> Path:
    if path is None:
        slug = make_slug(note.title, note.id)
        path = directory / f"{slug}.md"
    write_md(path, note.to_frontmatter_meta(), note.body)
    return path


def _enrich(note: Note, workspace: str, path: Path) -> dict:
    d = note.to_dict()
    d["workspace"] = workspace
    d["path"] = str(path)
    return d



from __future__ import annotations

# standard
import sys
import json
import typing
import inspect
import types as _builtin_types

# third party
# custom
from sunwaee.modules.gen.engine.types import Tool


def _is_union(origin: typing.Any) -> bool:
    """Check if is union."""

    if origin is typing.Union:
        return True
    if sys.version_info >= (3, 10):
        return origin is _builtin_types.UnionType
    return False  # pragma: no cover


def _unwrap_optional(annotation: typing.Any) -> tuple[typing.Any, bool]:
    """Strip `None` from a union. Returns (inner_type, is_optional)."""

    origin = typing.get_origin(annotation)
    if _is_union(origin):
        args = [a for a in typing.get_args(annotation) if a is not type(None)]
        if len(args) == 1:
            return args[0], True
    return annotation, False


def _py_to_jsonschema(annotation: typing.Any) -> dict[str, typing.Any]:
    """Transform python prop to json."""

    origin = typing.get_origin(annotation)
    args = typing.get_args(annotation)

    # Annotated[T, "description", ...]
    if origin is typing.Annotated:
        schema = _py_to_jsonschema(args[0])
        for meta in args[1:]:
            if isinstance(meta, str):
                schema["description"] = meta
        return schema

    # Union / Optional --> unwrap None, recurse on inner type
    if _is_union(origin):
        inner, _ = _unwrap_optional(annotation)
        return _py_to_jsonschema(inner)

    # Literal["a", "b"]
    if origin is typing.Literal:
        return {"type": "string", "enum": list(args)}

    # list[T]
    if origin is list:
        item_schema = _py_to_jsonschema(args[0]) if args else {"type": "string"}
        return {"type": "array", "items": item_schema}

    _PRIMITIVES = {str: "string", int: "integer", float: "number", bool: "boolean"}
    if annotation in _PRIMITIVES:
        return {"type": _PRIMITIVES[annotation]}

    return {"type": "string"}


def _is_optional_annotation(annotation: typing.Any) -> bool:
    """Return True if the annotation allows None (Optional / X | None)."""

    # unwrap Annotated first
    if typing.get_origin(annotation) is typing.Annotated:
        annotation = typing.get_args(annotation)[0]
    _, is_optional = _unwrap_optional(annotation)
    return is_optional


###  DECORATOR


def tool(description: str) -> typing.Callable:
    """Decorator that attaches a Tool definition to the function as `fn._tool`."""

    def decorator(fn: typing.Callable) -> typing.Callable:
        hints = typing.get_type_hints(fn, include_extras=True)
        sig = inspect.signature(fn)

        properties: dict[str, typing.Any] = {}
        required: list[str] = []

        for name, param in sig.parameters.items():
            annotation = hints.get(name, str)
            properties[name] = _py_to_jsonschema(annotation)

            has_default = param.default is not inspect.Parameter.empty
            if not has_default and not _is_optional_annotation(annotation):
                required.append(name)

        fn._tool = Tool(
            name=fn.__name__,
            description=description,
            parameters={
                "type": "object",
                "properties": properties,
                "required": required,
            },
            fn=fn,
        )
        return fn

    return decorator


### HELPERS


def ok(data: typing.Any, git_diff: str | None = None) -> str:
    """Successful tool call output."""

    result: dict[str, typing.Any] = {"ok": True, "data": data}
    if git_diff:
        result["git_diff"] = git_diff
    return json.dumps(result, default=str)


def err(message: str) -> str:
    """Error tool call output."""

    return json.dumps({"ok": False, "error": message})



# standard
from pathlib import Path

# third party
from rich import box
from rich.table import Table

# custom
from sunwaee.config import get_default_workspace, get_workspaces_base_dir
from sunwaee.core.audit import append_log
from sunwaee.core.fs import read_md
from sunwaee.core.output import console, success
from sunwaee.modules.workspaces.model import Workspace


def _load_workspace(workspace_dir: Path) -> Workspace:
    meta_path = workspace_dir / "metadata.md"
    if meta_path.exists():
        meta, body = read_md(meta_path)
        return Workspace.from_dict(workspace_dir.name, meta, body)
    return Workspace(name=workspace_dir.name)


def list_workspaces() -> None:
    """List all workspaces."""
    base = get_workspaces_base_dir()
    default = get_default_workspace()

    if not base.exists():
        success([], human_fn=lambda: console.print("[dim]No workspaces found.[/dim]"))
        return

    workspace_dirs = sorted(
        [d for d in base.iterdir() if d.is_dir() and not d.name.startswith(".")],
        key=lambda d: d.name,
    )

    workspaces = [_load_workspace(d) for d in workspace_dirs]

    data = [
        {**ws.to_dict(), "default": ws.name == default, "path": str(workspace_dirs[i])}
        for i, ws in enumerate(workspaces)
    ]

    def human_output() -> None:  # pragma: no cover
        if not workspaces:
            console.print("[dim]No workspaces found.[/dim]")
            return
        table = Table(box=box.SIMPLE, show_header=True, header_style="bold")
        table.add_column("Name")
        table.add_column("Default", width=10)
        table.add_column("Path", style="dim")
        for ws, d in zip(workspaces, workspace_dirs):
            is_default = ws.name == default
            table.add_row(
                f"[bold]{ws.name}[/bold]" if is_default else ws.name,
                "[green]✓[/green]" if is_default else "",
                str(d),
            )
        console.print(table)

    append_log(default or "personal", "workspaces", "list")
    success(data, human_fn=human_output)



# standard
from typing import Optional

# third party
import typer
from rich import box
from rich.table import Table

# custom
from sunwaee.config import get_default_workspace
from sunwaee.core.audit import append_log
from sunwaee.core.fs import list_md
from sunwaee.core.output import console, error, success
from ._helpers import _enrich, _load, _resolve_dirs

_NOTE_SORT_KEYS = {
    "updated": lambda n: n.updated_at,
    "created": lambda n: n.created_at,
    "title": lambda n: n.title.lower(),
}


def list_notes(
    tags: Optional[str] = typer.Option(
        None, "--tags", help="Filter by tag (comma-separated)"
    ),
    query: Optional[str] = typer.Option(
        None, "--query", "-q", help="Search term (matches title, tags, body)"
    ),
    sort: str = typer.Option(
        "updated", "--sort", help="Sort by: updated (default), created, title"
    ),
    workspace: Optional[str] = typer.Option(
        None, "--workspace", "-w", help="Workspace name (default: all workspaces)"
    ),
) -> None:
    """List all notes."""
    if sort not in _NOTE_SORT_KEYS:
        error(
            f"Invalid --sort value: {sort!r}. Choose from: {', '.join(_NOTE_SORT_KEYS)}",
            "VALIDATION_ERROR",
        )

    dirs = _resolve_dirs(workspace)
    multi_ws = workspace is None

    triples: list[tuple, ...] = []
    for ws, directory in dirs:
        for p in list_md(directory):
            triples.append((_load(p), p, ws))

    if tags:
        filter_tags = {t.strip().lower() for t in tags.split(",")}
        triples = [
            (n, p, w)
            for n, p, w in triples
            if any(t.lower() in filter_tags for t in n.tags)
        ]

    if query:
        q = query.lower()
        triples = [
            (n, p, w)
            for n, p, w in triples
            if q in n.title.lower()
            or q in n.body.lower()
            or any(q in t.lower() for t in n.tags)
        ]

    key_fn = _NOTE_SORT_KEYS[sort]
    reverse = sort != "title"
    triples = sorted(triples, key=lambda x: key_fn(x[0]), reverse=reverse)

    notes = [n for n, p, w in triples]
    paths = [p for n, p, w in triples]
    ws_names = [w for n, p, w in triples]

    _log_args: dict = {"results": len(notes)}
    if tags:
        _log_args["tags"] = tags
    if query:
        _log_args["query"] = query
    if sort != "updated":
        _log_args["sort"] = sort
    append_log(workspace or get_default_workspace(), "notes", "list", args=_log_args)

    def human_output() -> None:  # pragma: no cover
        if not notes:
            scope = "all workspaces" if multi_ws else f"workspace '{workspace}'"
            console.print(f"[dim]No notes in {scope}.[/dim]")
            return
        table = Table(box=box.SIMPLE, show_header=True, header_style="bold")
        table.add_column("ID", style="dim", width=10, no_wrap=True)
        table.add_column("Title")
        if multi_ws:
            table.add_column("Workspace", style="dim", width=14)
        table.add_column("Tags")
        table.add_column("Updated", style="dim", width=12)
        table.add_column("Path", style="dim")
        for note, path, ws_name in zip(notes, paths, ws_names):
            row = [note.id[:8], note.title]
            if multi_ws:
                row.append(ws_name)
            row += [
                ", ".join(note.tags) if note.tags else "—",
                note.updated_at[:10],
                str(path.resolve()),
            ]
            table.add_row(*row)
        console.print(table)

    success(
        [_enrich(n, ws_n, p) for n, p, ws_n in zip(notes, paths, ws_names)],
        human_fn=human_output,
    )



# standard
from typing import Optional

# third party
import typer
from rich import box
from rich.panel import Panel

# custom
from sunwaee.core.audit import append_log
from sunwaee.core.logger import get_logger
from sunwaee.core.output import console, error, success
from ._helpers import _enrich, _find_in_all, _load

_log = get_logger("modules.notes")


def read_note(
    query: str = typer.Argument(..., help="Note id, slug, or title"),
    workspace: Optional[str] = typer.Option(
        None,
        "--workspace",
        "-w",
        help="Workspace name (default: search all workspaces)",
    ),
) -> None:
    """Read a note."""
    found = _find_in_all(query, workspace)
    if found is None:
        _log.warning("Note not found: %r", query)
        error(f"Note not found: {query!r}", "NOT_FOUND")
    ws, directory, path = found

    note = _load(path)
    _log.debug("Reading note %r (id=%s)", note.title, note.id)
    append_log(ws, "notes", "read", item_id=note.id, item_title=note.title)

    def human_output() -> None:  # pragma: no cover
        from rich.markdown import Markdown

        meta_line = "\n".join(
            [
                f"[dim]id:[/dim] {note.id[:8]}  "
                f"[dim]workspace:[/dim] {ws}  "
                f"[dim]tags:[/dim] {', '.join(note.tags) if note.tags else '—'}  "
                f"[dim]updated:[/dim] {note.updated_at[:10]}",
                f"[dim]path:[/dim] {path}",
            ]
        )
        console.print(
            Panel(
                meta_line,
                title=f"[bold]{note.title}[/bold]",
                box=box.ROUNDED,
                expand=False,
            )
        )
        if note.body:
            console.print(Markdown(note.body))

    success(_enrich(note, ws, path), human_fn=human_output)



# standard
from typing import Optional

# third party
import typer

# custom
from sunwaee.config import is_machine_caller
from sunwaee.core.audit import append_log
from sunwaee.core.editor import open_editor
from sunwaee.core.logger import get_logger
from sunwaee.core.output import console, success
from sunwaee.modules.notes.model import Note
from ._helpers import _enrich, _resolve_dir, _save

_log = get_logger("modules.notes")


def create(
    title: str = typer.Option(..., "--title", "-t", help="Note title"),
    tags: Optional[str] = typer.Option(None, "--tags", help="Comma-separated tags"),
    body: Optional[str] = typer.Option(
        None, "--body", "-b", help="Note body (markdown)"
    ),
    workspace: Optional[str] = typer.Option(
        None, "--workspace", "-w", help="Workspace name (default: from config)"
    ),
) -> None:
    """Create a new note."""
    ws, directory = _resolve_dir(workspace)
    tag_list = [t.strip() for t in tags.split(",")] if tags else []

    if body is None and not is_machine_caller():  # pragma: no cover
        body = open_editor()

    note = Note(title=title, tags=tag_list, body=body or "")
    path = _save(note, directory)
    _log.info("Created note %r (id=%s) at %s", title, note.id, path)
    append_log(ws, "notes", "create", item_id=note.id, item_title=note.title)

    def _created() -> None:  # pragma: no cover
        console.print(f"[green]Created[/green] note '[bold]{note.title}[/bold]'")
        console.print(f"[dim]{path}[/dim]")

    success(_enrich(note, ws, path), human_fn=_created)



# standard
from typing import Annotated, Literal

# third party
# custom
from sunwaee.config import (
    get_active_project,
    get_all_projects,
    get_default_workspace,
    get_project_dir,
    get_workspace_dir,
    set_default_project,
)
from sunwaee.core.audit import append_log
from sunwaee.core.fs import read_md, write_md
from sunwaee.core.git import ensure_git_repo
from sunwaee.core.tools import err, ok, tool
from sunwaee.modules.projects.model import Project


def _load(ws: str, name: str) -> Project | None:
    meta_path = get_project_dir(ws, name) / "metadata.md"
    if not meta_path.exists():
        return None
    meta, body = read_md(meta_path)
    return Project.from_dict(name, meta, body)


@tool(
    "Manage projects. Actions: "
    "'create' (requires name, optional workspace and body), "
    "'list' returns all projects in the given workspace with full metadata, "
    "'read' (requires name, optional workspace) returns a single project in full detail, "
    "'update' (requires name; optional workspace, body, set_default)."
)
def projects(
    action: Annotated[Literal["create", "list", "read", "update"], "Action to perform"],
    name: Annotated[str | None, "Project name — required for create/update"] = None,
    workspace: Annotated[str | None, "Workspace name (uses default if omitted)"] = None,
    body: Annotated[str | None, "Project description / details (markdown)"] = None,
    set_default: Annotated[bool, "Set as the default project (update only)"] = False,
) -> str:
    ws = workspace or get_default_workspace()

    if action == "list":
        default = get_active_project(ws)
        result = []
        for project_name in get_all_projects(ws):
            proj = _load(ws, project_name) or Project(name=project_name)
            entry = proj.to_dict()
            entry["workspace"] = ws
            entry["default"] = proj.name == default
            result.append(entry)
        return ok(result)

    if not name:
        return err(f"'name' is required for action={action!r}")

    if action == "create":
        project_dir = get_project_dir(ws, name)
        if project_dir.exists():
            return err(f"Project already exists: {name!r}")
        project_dir.mkdir(parents=True, exist_ok=True)
        ensure_git_repo(get_workspace_dir(ws))
        proj = Project(name=name, body=body or "")
        write_md(project_dir / "metadata.md", proj.to_frontmatter_meta(), proj.body)
        append_log(ws, "projects", "create", item_id=proj.id, item_title=name)
        data = proj.to_dict()
        data["workspace"] = ws
        data["path"] = str(project_dir)
        return ok(data)

    if action == "read":
        project_dir = get_project_dir(ws, name)
        if not project_dir.exists():
            return err(f"Project not found: {name!r}")
        proj = _load(ws, name) or Project(name=name)
        append_log(ws, "projects", "read", item_id=proj.id, item_title=name)
        data = proj.to_dict()
        data["workspace"] = ws
        data["default"] = get_active_project(ws) == name
        return ok(data)

    if action == "update":
        project_dir = get_project_dir(ws, name)
        if not project_dir.exists():
            return err(f"Project not found: {name!r}")
        if body is None and not set_default:
            return err("Specify at least one of body or set_default=true")
        proj = _load(ws, name) or Project(name=name)
        if body is not None:
            proj.body = body
        if set_default:
            set_default_project(ws, name)
        proj.touch()
        write_md(project_dir / "metadata.md", proj.to_frontmatter_meta(), proj.body)
        append_log(ws, "projects", "update", item_id=proj.id, item_title=name)
        data = proj.to_dict()
        data["workspace"] = ws
        data["default"] = get_active_project(ws) == name
        return ok(data)

    return err(f"Unknown action: {action!r}")  # pragma: no cover


TOOLS = [projects._tool]



# third party
import typer

# custom
from .list import list_logs
from .read import read_log

app = typer.Typer(name="logs", help="View and search activity logs.", no_args_is_help=True)
app.command("list")(list_logs)
app.command("read")(read_log)



# standard
import os
import logging

# third party
# custom

_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


def get_logger(name: str) -> logging.Logger:
    """Return a logger scoped to *name* under the 'sunwaee' hierarchy."""

    return logging.getLogger(f"sunwaee.{name}")


def _configure() -> None:
    """Initialize logger."""

    raw = os.environ.get("SUNWAEE_LOG_LEVEL", "").lower()
    level = _LEVELS.get(raw, logging.WARNING)

    root = logging.getLogger("sunwaee")
    if root.handlers:
        return

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s: %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S",
        )
    )
    root.addHandler(handler)
    root.setLevel(level)


_configure()



# standard
import json

# third party
import pytest

# custom
from sunwaee.modules.gen.agent import stream_run
from sunwaee.modules.gen.engine.base import BaseEngine
from sunwaee.modules.gen.engine.types import (
    Message,
    Performance,
    Response,
    Role,
    StopReason,
    Tool,
    ToolCall,
    Usage,
)

### MOCK


class MockStreamEngine(BaseEngine):
    """Yields pre-configured Response chunk sequences, one list per stream() call."""

    def __init__(self, sequences: list[list[Response]]):
        self._sequences = list(sequences)
        self.calls: list[tuple[list[Message], list[Tool] | None]] = []

    async def chat(self, messages, tools=None) -> Response:
        raise NotImplementedError

    async def stream(self, messages, tools=None):
        self.calls.append((list(messages), tools))
        for chunk in self._sequences.pop(0):
            yield chunk


### HELPERS


def _final(
    stop_reason: StopReason = StopReason.END_TURN,
    tool_calls: list[ToolCall] | None = None,
) -> Response:
    """Final chunk — carries stop_reason, usage, performance."""
    return Response(
        streaming=True,
        stop_reason=stop_reason,
        tool_calls=tool_calls,
        usage=Usage(input_tokens=5, output_tokens=10, total_tokens=15),
        performance=Performance(latency=0.1, total_duration=0.5),
    )


def _text(t: str) -> Response:
    return Response(streaming=True, content=t)


def _reasoning(t: str) -> Response:
    return Response(streaming=True, reasoning_content=t)


def _sentinel(t: str) -> Response:
    return Response(streaming=True, synthetic=True, reasoning_content=t)


def _signature(sig: str) -> Response:
    return Response(streaming=True, reasoning_signature=sig)


def _tool_chunk(tc: ToolCall) -> Response:
    return Response(streaming=True, tool_calls=[tc])


def echo_tool(message: str = "pong") -> str:
    return json.dumps({"ok": True, "data": message})


ECHO = Tool(name="echo", description="Echo.", parameters={}, fn=echo_tool)


### TESTS


@pytest.mark.asyncio
async def test_simple_stream_no_tools():
    engine = MockStreamEngine([[_text("Hello"), _text(", world!"), _final()]])
    chunks = []
    async for chunk in stream_run([Message(role=Role.USER, content="Hi")], [], engine):
        chunks.append(chunk)

    assert len(engine.calls) == 1
    text_chunks = [c for c in chunks if c.content]
    assert "".join(c.content for c in text_chunks) == "Hello, world!"


@pytest.mark.asyncio
async def test_new_messages_populated():
    engine = MockStreamEngine([[_text("Answer"), _final()]])
    new_msgs: list[Message] = []
    async for _ in stream_run(
        [Message(role=Role.USER, content="Q")], [], engine, new_messages=new_msgs
    ):
        pass

    assert len(new_msgs) == 1
    assert new_msgs[0].role == Role.ASSISTANT
    assert new_msgs[0].content == "Answer"


@pytest.mark.asyncio
async def test_single_tool_call():
    tc = ToolCall(id="c1", name="echo", arguments={"message": "hi"})
    engine = MockStreamEngine(
        [
            [_tool_chunk(tc), _final(StopReason.TOOL_USE)],
            [_text("Done!"), _final()],
        ]
    )
    new_msgs: list[Message] = []
    chunks = []
    async for chunk in stream_run(
        [Message(role=Role.USER, content="Call echo")],
        [ECHO],
        engine,
        new_messages=new_msgs,
    ):
        chunks.append(chunk)

    assert len(engine.calls) == 2

    roles = [m.role.value for m in new_msgs]
    assert roles == ["assistant", "tool", "assistant"]

    tool_msg = next(m for m in new_msgs if m.role == Role.TOOL)
    assert tool_msg.tool_call_id == "c1"
    assert tool_msg.content is not None
    assert json.loads(tool_msg.content)["data"] == "hi"

    text_chunks = [c for c in chunks if c.content]
    assert "".join(c.content for c in text_chunks) == "Done!"


@pytest.mark.asyncio
async def test_multiple_tool_iterations():
    tc1 = ToolCall(id="c1", name="echo", arguments={"message": "first"})
    tc2 = ToolCall(id="c2", name="echo", arguments={"message": "second"})
    engine = MockStreamEngine(
        [
            [_tool_chunk(tc1), _final(StopReason.TOOL_USE)],
            [_tool_chunk(tc2), _final(StopReason.TOOL_USE)],
            [_text("All done"), _final()],
        ]
    )
    new_msgs: list[Message] = []
    async for _ in stream_run(
        [Message(role=Role.USER, content="Go")],
        [ECHO],
        engine,
        new_messages=new_msgs,
    ):
        pass

    assert len(engine.calls) == 3
    tool_msgs = [m for m in new_msgs if m.role == Role.TOOL]
    assert len(tool_msgs) == 2


@pytest.mark.asyncio
async def test_concurrent_tool_calls():
    second = Tool(
        name="second",
        description=".",
        parameters={},
        fn=lambda: json.dumps({"ok": True, "data": "s"}),
    )
    tc1 = ToolCall(id="c1", name="echo", arguments={"message": "a"})
    tc2 = ToolCall(id="c2", name="second", arguments={})
    engine = MockStreamEngine(
        [
            [_tool_chunk(tc1), _tool_chunk(tc2), _final(StopReason.TOOL_USE)],
            [_text("Done"), _final()],
        ]
    )
    new_msgs: list[Message] = []
    async for _ in stream_run(
        [Message(role=Role.USER, content="Both")],
        [ECHO, second],
        engine,
        new_messages=new_msgs,
    ):
        pass

    tool_msgs = [m for m in new_msgs if m.role == Role.TOOL]
    assert len(tool_msgs) == 2
    assert {m.tool_call_id for m in tool_msgs} == {"c1", "c2"}


@pytest.mark.asyncio
async def test_reasoning_chunks_forwarded():
    engine = MockStreamEngine(
        [[_reasoning("think..."), _signature("sig-1"), _text("Answer"), _final()]]
    )
    chunks = []
    async for chunk in stream_run([Message(role=Role.USER, content="Q")], [], engine):
        chunks.append(chunk)

    reasoning = "".join(c.reasoning_content for c in chunks if c.reasoning_content)
    assert reasoning == "think..."
    sig_chunks = [c for c in chunks if c.reasoning_signature]
    assert sig_chunks[-1].reasoning_signature == "sig-1"


@pytest.mark.asyncio
async def test_reasoning_stored_in_new_messages():
    engine = MockStreamEngine(
        [[_reasoning("chain"), _signature("sig-x"), _text("Result"), _final()]]
    )
    new_msgs: list[Message] = []
    async for _ in stream_run(
        [Message(role=Role.USER, content="Q")], [], engine, new_messages=new_msgs
    ):
        pass

    assistant = new_msgs[0]
    assert assistant.reasoning_content == "chain"
    assert assistant.reasoning_signature == "sig-x"


@pytest.mark.asyncio
async def test_unknown_tool_returns_error():
    tc = ToolCall(id="c1", name="ghost", arguments={})
    engine = MockStreamEngine(
        [
            [_tool_chunk(tc), _final(StopReason.TOOL_USE)],
            [_text("Ok"), _final()],
        ]
    )
    new_msgs: list[Message] = []
    async for _ in stream_run(
        [Message(role=Role.USER, content="Hi")], [ECHO], engine, new_messages=new_msgs
    ):
        pass

    tool_msg = next(m for m in new_msgs if m.role == Role.TOOL)
    assert tool_msg.content is not None
    data = json.loads(tool_msg.content)
    assert data["ok"] is False
    assert "ghost" in data["error"]


@pytest.mark.asyncio
async def test_max_iterations():
    tc = ToolCall(id="c1", name="echo", arguments={"message": "x"})
    engine = MockStreamEngine([[_tool_chunk(tc), _final(StopReason.TOOL_USE)]] * 3)
    chunks = []
    async for chunk in stream_run(
        [Message(role=Role.USER, content="Hi")],
        [ECHO],
        engine,
        max_iterations=2,
    ):
        chunks.append(chunk)

    assert len(engine.calls) == 2


@pytest.mark.asyncio
async def test_history_not_mutated():
    engine = MockStreamEngine([[_text("Hi"), _final()]])
    original = [Message(role=Role.USER, content="Hello")]
    original_len = len(original)

    async for _ in stream_run(original, [], engine):
        pass

    assert len(original) == original_len


@pytest.mark.asyncio
async def test_new_messages_excludes_input():
    engine = MockStreamEngine([[_text("Response"), _final()]])
    history = [
        Message(role=Role.USER, content="Old"),
        Message(role=Role.ASSISTANT, content="Old reply"),
        Message(role=Role.USER, content="New"),
    ]
    new_msgs: list[Message] = []
    async for _ in stream_run(history, [], engine, new_messages=new_msgs):
        pass

    assert len(new_msgs) == 1
    assert new_msgs[0].content == "Response"


@pytest.mark.asyncio
async def test_no_new_messages_arg_works():
    """stream_run works fine without passing new_messages."""
    engine = MockStreamEngine([[_text("Ok"), _final()]])
    chunks = []
    async for chunk in stream_run([Message(role=Role.USER, content="Hi")], [], engine):
        chunks.append(chunk)
    assert any(c.content for c in chunks)


@pytest.mark.asyncio
async def test_synthetic_sentinel_yielded_but_not_stored():
    """Synthetic sentinel chunks are forwarded to the caller but not accumulated
    into the assistant message, so they are never sent back to the provider API."""
    engine = MockStreamEngine(
        [
            [
                _sentinel(
                    "Reasoning in progress - tokens not available for this model"
                ),
                _text("Answer"),
                _final(),
            ]
        ]
    )
    new_msgs: list[Message] = []
    chunks = []
    async for chunk in stream_run(
        [Message(role=Role.USER, content="Q")], [], engine, new_messages=new_msgs
    ):
        chunks.append(chunk)

    # sentinel was yielded to the caller
    sentinel_chunks = [c for c in chunks if c.synthetic]
    assert len(sentinel_chunks) == 1

    # but not stored in the assistant message history
    assistant = new_msgs[0]
    assert assistant.reasoning_content is None



# standard
from pathlib import Path
from typing import Optional

# custom
from sunwaee.modules.tasks.model import Task
from sunwaee.config import (
    get_all_workspaces,
    get_default_workspace,
    get_module_dir,
)
from sunwaee.core.fs import find_md, list_md, make_slug, read_md, write_md


def _resolve_dir(workspace: Optional[str]) -> tuple[str, Path]:
    workspace = workspace or get_default_workspace()
    return workspace, get_module_dir(workspace, "tasks")


def _resolve_dirs(workspace: Optional[str]) -> list[tuple[str, Path]]:
    if workspace:
        return [_resolve_dir(workspace)]
    return [_resolve_dir(ws) for ws in get_all_workspaces()]


def _find_in_all(
    query: str, workspace: Optional[str]
) -> Optional[tuple[str, Path, Path]]:
    """Search for a task across workspaces. Returns (ws_name, directory, path) or None."""
    for ws, directory in _resolve_dirs(workspace):
        path = find_md(query, directory)
        if path is not None:
            return ws, directory, path
    return None


def _load(path: Path) -> Task:
    meta, body = read_md(path)
    return Task.from_dict(meta, body)


def _save(task: Task, directory: Path, path: Optional[Path] = None) -> Path:
    if path is None:
        slug = make_slug(task.title, task.id)
        path = directory / f"{slug}.md"
    write_md(path, task.to_frontmatter_meta(), task.body)
    return path


def _enrich(task: Task, workspace: str, path: Path) -> dict:
    d = task.to_dict()
    d["workspace"] = workspace
    d["path"] = str(path)
    return d


def _find_children(parent_id: str, directory: Path) -> list[tuple[Task, Path]]:
    results = []
    for p in list_md(directory):
        task = _load(p)
        if task.parent_id == parent_id:
            results.append((task, p))
    return results


def _delete_cascade(task: Task, path: Path, directory: Path) -> list[tuple[Task, Path]]:
    deleted = []
    for child, child_path in _find_children(task.id, directory):
        deleted.extend(_delete_cascade(child, child_path, directory))
    path.unlink()
    deleted.append((task, path))
    return deleted



from __future__ import annotations

# standard
# third party
# custom
from ..model import Model

MODELS: list[Model] = [
    Model(
        name="claude-sonnet-4-6",
        display_name="Claude Sonnet 4.6",
        provider="anthropic",
        context_window=1_000_000,
        max_output_tokens=128_000,
        input_price_per_mtok=3.0,
        output_price_per_mtok=15.0,
        cache_read_price_per_mtok=0.3,
        cache_write_price_per_mtok=3.75,
        input_price_per_mtok_200k=6,
        output_price_per_mtok_200k=22.5,
        cache_read_price_per_mtok_200k=0.6,
        cache_write_price_per_mtok_200k=7.5,
        supports_vision=True,
        supports_tools=True,
        supports_thinking=True,
        supports_reasoning_tokens=True,
        release_date="2026-02-17",
    ),
    Model(
        name="claude-opus-4-6",
        display_name="Claude Opus 4.6",
        provider="anthropic",
        context_window=1_000_000,
        max_output_tokens=128_000,
        input_price_per_mtok=5.0,
        output_price_per_mtok=25.0,
        cache_read_price_per_mtok=0.5,
        cache_write_price_per_mtok=6.25,
        input_price_per_mtok_200k=10,
        output_price_per_mtok_200k=37.5,
        cache_read_price_per_mtok_200k=1,
        cache_write_price_per_mtok_200k=12.5,
        supports_vision=True,
        supports_tools=True,
        supports_thinking=True,
        supports_reasoning_tokens=True,
        release_date="2026-02-04",
    ),
    Model(
        name="claude-opus-4-5",
        display_name="Claude Opus 4.5",
        provider="anthropic",
        context_window=200_000,
        max_output_tokens=64_000,
        input_price_per_mtok=5.0,
        output_price_per_mtok=25.0,
        cache_read_price_per_mtok=0.5,
        cache_write_price_per_mtok=6.25,
        supports_vision=True,
        supports_tools=True,
        supports_thinking=True,
        supports_reasoning_tokens=True,
        release_date="2025-11-24",
    ),
    Model(
        name="claude-haiku-4-5",
        display_name="Claude Haiku 4.5",
        provider="anthropic",
        context_window=200_000,
        max_output_tokens=64_000,
        input_price_per_mtok=1.0,
        output_price_per_mtok=5.0,
        cache_read_price_per_mtok=0.1,
        cache_write_price_per_mtok=1.25,
        supports_vision=True,
        supports_tools=True,
        supports_thinking=True,
        supports_reasoning_tokens=True,
        release_date="2025-10-15",
    ),
    Model(
        name="claude-sonnet-4-5",
        display_name="Claude Sonnet 4.5",
        provider="anthropic",
        context_window=1_000_000,
        max_output_tokens=64_000,
        input_price_per_mtok=3.0,
        output_price_per_mtok=15.0,
        cache_read_price_per_mtok=0.3,
        cache_write_price_per_mtok=3.75,
        input_price_per_mtok_200k=6.0,
        output_price_per_mtok_200k=22.5,
        cache_read_price_per_mtok_200k=0.6,
        cache_write_price_per_mtok_200k=7.5,
        supports_vision=True,
        supports_tools=True,
        supports_thinking=True,
        supports_reasoning_tokens=True,
        release_date="2025-09-29",
    ),
]



# standard
import json

# third party
# custom
from sunwaee.modules.notes.tools import notes, TOOLS as NOTE_TOOLS

WS = "personal"


def test_notes_tool_create(ws):
    data = json.loads(notes(action="create", title="My note", workspace=WS))
    assert data["ok"] is True
    assert data["data"]["title"] == "My note"


def test_notes_tool_create_with_body_and_tags(ws):
    data = json.loads(notes(action="create", title="Tagged", body="content", tags=["a", "b"], workspace=WS))
    assert data["data"]["tags"] == ["a", "b"]
    assert data["data"]["body"] == "content"


def test_notes_tool_create_missing_title(ws):
    data = json.loads(notes(action="create", workspace=WS))
    assert data["ok"] is False
    assert "title" in data["error"]


def test_notes_tool_list_empty(ws):
    data = json.loads(notes(action="list", workspace=WS))
    assert data["ok"] is True
    assert data["data"] == []


def test_notes_tool_list(ws):
    notes(action="create", title="Note A", workspace=WS)
    notes(action="create", title="Note B", workspace=WS)
    data = json.loads(notes(action="list", workspace=WS))
    assert len(data["data"]) == 2


def test_notes_tool_list_filter_by_tags(ws):
    notes(action="create", title="Tagged", tags=["dev"], workspace=WS)
    notes(action="create", title="Other", tags=["other"], workspace=WS)
    data = json.loads(notes(action="list", tags=["dev"], workspace=WS))
    assert len(data["data"]) == 1
    assert data["data"][0]["title"] == "Tagged"


def test_notes_tool_list_filter_by_query(ws):
    notes(action="create", title="Find me", workspace=WS)
    notes(action="create", title="Skip", workspace=WS)
    data = json.loads(notes(action="list", query="Find", workspace=WS))
    assert any(n["title"] == "Find me" for n in data["data"])


def test_notes_tool_read(ws):
    notes(action="create", title="Show me", body="details", workspace=WS)
    data = json.loads(notes(action="read", query="Show me", workspace=WS))
    assert data["ok"] is True
    assert data["data"]["body"] == "details"


def test_notes_tool_read_not_found(ws):
    data = json.loads(notes(action="read", query="ghost", workspace=WS))
    assert data["ok"] is False


def test_notes_tool_read_missing_query(ws):
    data = json.loads(notes(action="read", workspace=WS))
    assert data["ok"] is False
    assert "query" in data["error"]


def test_notes_tool_update(ws):
    notes(action="create", title="Edit me", workspace=WS)
    data = json.loads(notes(action="update", query="Edit me", body="updated", workspace=WS))
    assert data["ok"] is True
    assert data["data"]["body"] == "updated"


def test_notes_tool_update_not_found(ws):
    data = json.loads(notes(action="update", query="ghost", title="X", workspace=WS))
    assert data["ok"] is False


def test_notes_tool_delete(ws):
    notes(action="create", title="Delete me", workspace=WS)
    data = json.loads(notes(action="delete", query="Delete me", workspace=WS))
    assert data["ok"] is True
    gone = json.loads(notes(action="read", query="Delete me", workspace=WS))
    assert gone["ok"] is False


def test_notes_tool_delete_not_found(ws):
    data = json.loads(notes(action="delete", query="ghost", workspace=WS))
    assert data["ok"] is False


def test_notes_tools_registry():
    assert len(NOTE_TOOLS) == 1
    assert NOTE_TOOLS[0].name == "notes"
    assert NOTE_TOOLS[0].fn is not None
    assert "create" in NOTE_TOOLS[0].parameters["properties"]["action"]["enum"]
    assert "read" in NOTE_TOOLS[0].parameters["properties"]["action"]["enum"]
    assert "update" in NOTE_TOOLS[0].parameters["properties"]["action"]["enum"]



# standard
import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

# third party
import typer
from rich import box
from rich.table import Table

# custom
from sunwaee.modules.logs.model import LogEntry
from sunwaee.config import (
    get_all_workspaces,
    get_default_workspace,
    get_workspaces_base_dir,
)
from sunwaee.core.audit import append_log
from sunwaee.core.output import console, error, success


def _logs_dir(workspace: str) -> Path:
    return get_workspaces_base_dir() / workspace / "logs"


def _resolve_workspaces(workspace: Optional[str]) -> list[str]:
    return [workspace] if workspace else get_all_workspaces()


def _read_day(workspace: str, date: str) -> list[LogEntry]:
    year, month, day = date.split("-")
    path = _logs_dir(workspace) / year / month / day / "logs.jsonl"
    if not path.exists():
        return []
    return _parse_jsonl(path)


def _read_month(workspace: str, month: str) -> list[LogEntry]:
    year, mon = month.split("-")
    month_dir = _logs_dir(workspace) / year / mon
    if not month_dir.exists():  # pragma: no cover
        return []
    entries = []
    for day_dir in sorted(month_dir.iterdir()):
        if day_dir.is_dir():
            log_file = day_dir / "logs.jsonl"
            if log_file.exists():
                entries.extend(_parse_jsonl(log_file))
    return entries


def _read_all(workspace: str) -> list[LogEntry]:
    base = _logs_dir(workspace)
    if not base.exists():  # pragma: no cover
        return []
    entries = []
    for log_file in sorted(base.rglob("logs.jsonl")):
        entries.extend(_parse_jsonl(log_file))
    return entries


def _parse_jsonl(path: Path) -> list[LogEntry]:
    entries = []
    with open(path) as f:
        for line in f:
            line = line.strip()
            if line:
                try:
                    entries.append(LogEntry.from_dict(json.loads(line)))
                except (json.JSONDecodeError, KeyError):  # pragma: no cover
                    pass
    return entries


def _collect(
    workspaces: list[str],
    date: Optional[str],
    month: Optional[str],
    all_time: bool,
) -> list[LogEntry]:
    entries: list[LogEntry] = []
    if all_time:
        for ws in workspaces:
            entries.extend(_read_all(ws))
    elif month:
        for ws in workspaces:
            entries.extend(_read_month(ws, month))
    else:
        target = date or datetime.now(timezone.utc).strftime("%Y-%m-%d")
        for ws in workspaces:
            entries.extend(_read_day(ws, target))
    return entries


def _validate_date_flags(
    date: Optional[str], month: Optional[str], all_time: bool
) -> None:
    if sum([bool(date), bool(month), all_time]) > 1:
        error("Use only one of --date, --month, or --all", "VALIDATION_ERROR")
    if month:
        try:
            datetime.strptime(month, "%Y-%m")
        except ValueError:
            error(f"Invalid --month format: {month!r}. Use YYYY-MM", "VALIDATION_ERROR")
    if date:
        try:
            datetime.strptime(date, "%Y-%m-%d")
        except ValueError:
            error(
                f"Invalid --date format: {date!r}. Use YYYY-MM-DD", "VALIDATION_ERROR"
            )


def _render_table(entries: list[LogEntry], multi_ws: bool) -> None:  # pragma: no cover
    if not entries:
        console.print("[dim]No logs found.[/dim]")
        return
    table = Table(box=box.SIMPLE, show_header=True, header_style="bold")
    table.add_column("Timestamp", style="dim", width=20, no_wrap=True)
    table.add_column("Caller", width=8)
    if multi_ws:
        table.add_column("Workspace", style="dim", width=14)
    table.add_column("Module", width=12)
    table.add_column("Operation", width=12)
    table.add_column("Item")
    table.add_column("OK", width=4)
    for e in entries:
        ts = e.timestamp[:19].replace("T", " ")
        ok_str = "[green]✓[/green]" if e.ok else "[red]✗[/red]"
        item = e.item_title or e.item_id or "—"
        row = [ts, e.caller]
        if multi_ws:
            row.append(e.workspace)
        row += [e.module, e.operation, item, ok_str]
        table.add_row(*row)
    console.print(table)


def list_logs(
    workspace: Optional[str] = typer.Option(
        None, "--workspace", "-w", help="Workspace name (default: all workspaces)"
    ),
    date: Optional[str] = typer.Option(
        None, "--date", help="Show logs for a specific day (YYYY-MM-DD, default: today)"
    ),
    month: Optional[str] = typer.Option(
        None, "--month", help="Show logs for a full month (YYYY-MM)"
    ),
    all_time: bool = typer.Option(
        False, "--all", help="Show all logs regardless of date"
    ),
    module: Optional[str] = typer.Option(
        None, "--module", "-m", help="Filter by module (notes, tasks, workspaces...)"
    ),
    operation: Optional[str] = typer.Option(
        None, "--operation", "-o", help="Filter by operation (create, edit, delete...)"
    ),
    caller: Optional[str] = typer.Option(
        None, "--caller", "-c", help="Filter by caller (human, api, sun)"
    ),
    query: Optional[str] = typer.Option(
        None,
        "--query",
        "-q",
        help="Search term (matches item title, id, module, operation, caller, workspace)",
    ),
    limit: int = typer.Option(
        50, "--limit", "-n", help="Maximum number of entries to show"
    ),
) -> None:
    """List activity logs. Defaults to today's logs."""
    _validate_date_flags(date, month, all_time)

    workspaces = _resolve_workspaces(workspace)
    multi_ws = workspace is None

    entries = _collect(workspaces, date, month, all_time)

    if module:
        entries = [e for e in entries if e.module == module]
    if operation:
        entries = [e for e in entries if e.operation == operation]
    if caller:
        entries = [e for e in entries if e.caller == caller]
    if query:
        entries = [e for e in entries if e.matches(query)]

    entries = sorted(entries, key=lambda e: e.timestamp, reverse=True)[:limit]

    _log_args: dict = {"results": len(entries)}
    if date:
        _log_args["date"] = date
    if month:
        _log_args["month"] = month
    if all_time:
        _log_args["all"] = True
    if module:
        _log_args["module"] = module
    if operation:
        _log_args["operation"] = operation
    if caller:
        _log_args["caller"] = caller
    if query:
        _log_args["query"] = query
    append_log(workspace or get_default_workspace(), "logs", "list", args=_log_args)

    success(
        [e.to_dict() for e in entries],
        human_fn=lambda: _render_table(entries, multi_ws),
    )



# standard
from pathlib import Path
from typing import Optional

# third party
import typer
from rich import box
from rich.table import Table

# custom
from sunwaee.config import get_active_project, get_all_projects, get_default_workspace, get_project_dir
from sunwaee.core.audit import append_log
from sunwaee.core.fs import read_md
from sunwaee.core.output import console, success
from sunwaee.modules.projects.model import Project


def _load_project(project_dir: Path) -> Project:
    meta_path = project_dir / "metadata.md"
    if meta_path.exists():
        meta, body = read_md(meta_path)
        return Project.from_dict(project_dir.name, meta, body)
    return Project(name=project_dir.name)


def list_projects(
    workspace: Optional[str] = typer.Option(
        None, "--workspace", "-w", help="Workspace name (default: from config)"
    ),
) -> None:
    """List all projects in a workspace."""
    ws = workspace or get_default_workspace()
    default = get_active_project(ws)
    project_names = get_all_projects(ws)

    projects = [_load_project(get_project_dir(ws, name)) for name in project_names]

    data = [
        {**proj.to_dict(), "workspace": ws, "default": proj.name == default}
        for proj in projects
    ]

    def human_output() -> None:  # pragma: no cover
        if not projects:
            console.print(f"[dim]No projects in workspace '{ws}'.[/dim]")
            return
        table = Table(box=box.SIMPLE, show_header=True, header_style="bold")
        table.add_column("Name")
        table.add_column("Default", width=10)
        for proj in projects:
            is_default = proj.name == default
            table.add_row(
                f"[bold]{proj.name}[/bold]" if is_default else proj.name,
                "[green]✓[/green]" if is_default else "",
            )
        console.print(table)

    append_log(ws, "projects", "list")
    success(data, human_fn=human_output)



# standard
# third party
# custom
from ...conftest import ok
from sunwaee.cli import app


def test_list_workspaces_empty(runner, env):
    data = ok(runner.invoke(app, ["workspaces", "list"]))
    assert data == []


def test_list_workspaces(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    runner.invoke(app, ["workspaces", "create", "pro"])
    data = ok(runner.invoke(app, ["workspaces", "list"]))
    names = [w["name"] for w in data]
    assert "personal" in names
    assert "pro" in names


def test_list_workspace_has_id(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    data = ok(runner.invoke(app, ["workspaces", "list"]))
    assert all("id" in w for w in data)
    assert all(len(w["id"]) == 36 for w in data)


def test_list_workspace_has_body(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    data = ok(runner.invoke(app, ["workspaces", "list"]))
    assert all("body" in w for w in data)


def test_list_shows_default(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    runner.invoke(app, ["workspaces", "create", "pro"])
    data = ok(runner.invoke(app, ["workspaces", "list"]))
    defaults = [w for w in data if w["default"]]
    assert len(defaults) == 1
    assert defaults[0]["name"] == "personal"


def test_list_shows_default_after_update(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    runner.invoke(app, ["workspaces", "create", "pro"])
    ok(runner.invoke(app, ["workspaces", "update", "pro", "--set-default"]))
    data = ok(runner.invoke(app, ["workspaces", "list"]))
    defaults = [w for w in data if w["default"]]
    assert defaults[0]["name"] == "pro"



# standard
import json

# third party
# custom
from sunwaee.modules.workspaces.tools import workspaces, TOOLS as WORKSPACE_TOOLS


def test_workspace_tool_list(env):
    data = json.loads(workspaces(action="list"))
    assert data["ok"] is True
    assert isinstance(data["data"], list)


def test_workspace_tool_list_has_metadata(env):
    workspaces(action="create", name="pro")
    data = json.loads(workspaces(action="list"))
    entry = data["data"][0]
    assert "id" in entry
    assert "body" in entry
    assert "default" in entry
    assert "path" in entry


def test_workspace_tool_create(env, tmp_path):
    data = json.loads(workspaces(action="create", name="pro"))
    assert data["ok"] is True
    assert data["data"]["name"] == "pro"
    assert "id" in data["data"]
    assert (tmp_path / "workspaces" / "pro").is_dir()


def test_workspace_tool_create_with_body(env):
    data = json.loads(workspaces(action="create", name="pro", body="My workspace"))
    assert data["ok"] is True
    assert data["data"]["body"] == "My workspace"


def test_workspace_tool_create_requires_name(env):
    data = json.loads(workspaces(action="create"))
    assert data["ok"] is False


def test_workspace_tool_create_duplicate_errors(env):
    workspaces(action="create", name="pro")
    data = json.loads(workspaces(action="create", name="pro"))
    assert data["ok"] is False


def test_workspace_tool_read(env):
    workspaces(action="create", name="pro", body="My workspace")
    data = json.loads(workspaces(action="read", name="pro"))
    assert data["ok"] is True
    assert data["data"]["name"] == "pro"
    assert data["data"]["body"] == "My workspace"
    assert "id" in data["data"]
    assert "default" in data["data"]
    assert "path" in data["data"]


def test_workspace_tool_read_requires_name(env):
    data = json.loads(workspaces(action="read"))
    assert data["ok"] is False


def test_workspace_tool_read_not_found_errors(env):
    data = json.loads(workspaces(action="read", name="ghost"))
    assert data["ok"] is False


def test_workspace_tool_update_body(env):
    workspaces(action="create", name="pro")
    data = json.loads(workspaces(action="update", name="pro", body="Updated"))
    assert data["ok"] is True
    assert data["data"]["body"] == "Updated"


def test_workspace_tool_update_set_default(env):
    workspaces(action="create", name="personal")
    workspaces(action="create", name="pro")
    data = json.loads(workspaces(action="update", name="pro", set_default=True))
    assert data["ok"] is True
    assert data["data"]["default"] is True


def test_workspace_tool_update_requires_name(env):
    data = json.loads(workspaces(action="update"))
    assert data["ok"] is False


def test_workspace_tool_update_no_changes_errors(env):
    workspaces(action="create", name="pro")
    data = json.loads(workspaces(action="update", name="pro"))
    assert data["ok"] is False


def test_workspace_tool_update_not_found_errors(env):
    data = json.loads(workspaces(action="update", name="ghost", body="x"))
    assert data["ok"] is False


def test_workspace_tools_registry():
    assert len(WORKSPACE_TOOLS) == 1
    assert WORKSPACE_TOOLS[0].name == "workspaces"
    assert WORKSPACE_TOOLS[0].fn is not None



from __future__ import annotations

# third party
import typer
from prompt_toolkit import PromptSession

# custom
from sunwaee.config import get_sun_default_model, set_sun_default_model
from sunwaee.core.output import console
from ..engine.models import list_models
from ..repl.input import ModelCompleter


def set_model() -> None:
    """Interactively select and save the default model."""

    models = list_models()
    model_names = [m.name for m in models]
    current_model, current_provider = get_sun_default_model()
    console.print(f"  Current: [bold]{current_model}[/bold]  {current_provider}")
    console.print("  Type a model name (Tab to autocomplete, Enter to confirm)\n")

    pt = PromptSession(
        completer=ModelCompleter(model_names),
        complete_while_typing=True,
    )
    try:
        chosen = pt.prompt("  model > ").strip()
    except (EOFError, KeyboardInterrupt):
        return

    if not chosen:
        return

    found = next((m for m in models if m.name == chosen), None)
    if not found:
        console.print(f"[red]Model {chosen!r} not found.[/red]")
        raise typer.Exit(1)

    set_sun_default_model(found.name, found.provider)
    console.print(
        f"  [green]Default model set to[/green] [bold]{found.name}[/bold]  {found.provider}"
    )



from __future__ import annotations

import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pathspec

from prompt_toolkit import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.layout import Layout
from prompt_toolkit.layout.containers import HSplit, Window, ConditionalContainer
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.filters import Condition, has_completions
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.menus import CompletionsMenuControl
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings
from prompt_toolkit.key_binding.bindings.emacs import load_emacs_bindings
from prompt_toolkit.styles import Style

STYLE = Style.from_dict(
    {
        "": "fg:#d7d7d7 bg:default",
        "completion-menu": "noinherit fg:#787d82 bg:default",
        "completion-menu.completion": "noinherit fg:#787d82 bg:default",
        "completion-menu.completion.current": "noinherit bold fg:#d7d7d7 bg:default",
        "completion-menu.meta": "noinherit fg:#787d82 bg:default",
        "completion-menu.meta.completion": "noinherit fg:#787d82 bg:default",
        "completion-menu.meta.completion.current": "noinherit fg:#d7d7d7 bg:default",
        "scrollbar.background": "bg:default",
        "scrollbar.button": "bg:default",
    }
)

_DISPLAY_MIN_WIDTH = 25
_DISPLAY_META_RESERVE = 32


def _fmt(text: str) -> str:
    """Pad text to min width and truncate with ellipsis if needed."""
    cols = shutil.get_terminal_size().columns
    max_width = max(_DISPLAY_MIN_WIDTH, cols - _DISPLAY_META_RESERVE)
    if len(text) > max_width:
        return text[: max_width - 1] + "…"
    return text.ljust(_DISPLAY_MIN_WIDTH)


REPL_COMMANDS: dict[str, str] = {
    "help": "show this help",
    "quit": "quit the REPL",
    "exit": "quit the REPL",
    "clear": "clear conversation history",
    "session": "show current session info",
    "cost": "show accumulated cost and token usage",
    "model": "show or switch current model",
}

_MAX_FILE_COMPLETIONS = 20
_gitignore_cache: tuple[Path, "pathspec.PathSpec"] | None = None  # noqa: F821


def _load_gitspec(cwd: Path) -> "pathspec.PathSpec":  # noqa: F821
    """Return a pathspec.PathSpec for the gitignore at cwd (cached)."""
    import pathspec  # noqa: PLC0415

    global _gitignore_cache
    if _gitignore_cache and _gitignore_cache[0] == cwd:
        return _gitignore_cache[1]

    lines: list[str] = []
    gitignore = cwd / ".gitignore"
    if gitignore.exists():
        lines = gitignore.read_text(errors="replace").splitlines()
    spec: pathspec.PathSpec = pathspec.PathSpec.from_lines("gitwildmatch", lines)
    _gitignore_cache = (cwd, spec)
    return spec


def _is_hidden(path: Path) -> bool:
    return any(part.startswith(".") for part in path.parts)


def _file_completions(partial: str):
    """Yield Completion entries for @<partial> filesystem matches.

    - No '/' in partial → recursive search for paths containing the string.
    - Partial ends with '/' → list that directory.
    - Otherwise → list parent dir filtered by last component prefix.
    Dotfiles and gitignored paths are always excluded.
    """

    cwd = Path.cwd()
    spec = _load_gitspec(cwd)

    def _ignored(p: Path) -> bool:
        rel = p.relative_to(cwd)
        rel_str = str(rel) + ("/" if p.is_dir() else "")
        return _is_hidden(rel) or bool(spec.match_file(rel_str))

    # ── recursive name search ─────────────────────────────────────────────
    if partial and "/" not in partial:
        needle = partial.lower()
        count = 0
        try:
            for dirpath, dirnames, filenames in os.walk(cwd):
                root = Path(dirpath)
                rel_root = root.relative_to(cwd)

                # prune ignored/hidden dirs in-place so os.walk won't descend
                dirnames[:] = sorted(
                    d
                    for d in dirnames
                    if not d.startswith(".")
                    and not spec.match_file(str(rel_root / d) + "/")
                )

                for name in sorted(filenames):
                    if name.startswith("."):
                        continue
                    rel = rel_root / name
                    rel_str = str(rel)
                    if spec.match_file(rel_str):
                        continue
                    if needle not in rel_str.lower():
                        continue
                    yield Completion(
                        rel_str, start_position=-len(partial), display=_fmt(rel_str)
                    )
                    count += 1
                    if count >= _MAX_FILE_COMPLETIONS:
                        return
        except (PermissionError, FileNotFoundError):
            pass
        return

    # ── directory listing ─────────────────────────────────────────────────
    base = Path(partial)
    if partial.endswith("/"):
        search_dir = cwd / base
        prefix = partial
        name_filter = ""
    else:
        search_dir = cwd / base.parent
        prefix = str(base.parent) + "/" if base.parent != Path(".") else ""
        name_filter = base.name

    try:
        entries = sorted(search_dir.iterdir(), key=lambda p: (p.is_file(), p.name))
    except (PermissionError, FileNotFoundError):
        return

    count = 0
    for entry in entries:
        if _ignored(entry):
            continue
        if name_filter and not entry.name.startswith(name_filter):
            continue
        display = prefix + entry.name + ("/" if entry.is_dir() else "")
        raw = f"+ {display}"
        yield Completion(raw, start_position=-len(partial), display=_fmt(raw))
        count += 1
        if count >= _MAX_FILE_COMPLETIONS:
            break


class ModelCompleter(Completer):
    """Completes bare model names (used by set-model)."""

    def __init__(self, model_names: list[str]) -> None:
        self._model_names = model_names

    def get_completions(self, document, complete_event):  # type: ignore[override]
        partial = document.text
        for name in self._model_names:
            if name.startswith(partial):
                yield Completion(name, start_position=-len(partial))


class ReplCompleter(Completer):
    """Completes REPL commands."""

    def __init__(self, model_names: list[str]) -> None:
        self._model_names = model_names

    def get_completions(self, document, complete_event):  # type: ignore[override]
        text_before = document.text_before_cursor

        # ── @file completion ──────────────────────────────────────────────────
        at_pos = text_before.rfind("@")
        if at_pos != -1:
            partial = text_before[at_pos + 1 :]
            if " " not in partial:
                yield from _file_completions(partial)
                return

        # ── :command completion ───────────────────────────────────────────────
        text = document.text
        if not text.startswith(":"):
            return
        after_colon = text[1:]
        parts = after_colon.split(" ", 1)

        if len(parts) == 1:
            partial = parts[0].lower()
            for cmd, desc in REPL_COMMANDS.items():
                if cmd.startswith(partial):
                    yield Completion(
                        cmd,
                        start_position=-len(partial),
                        display=_fmt(cmd),
                        display_meta=desc,
                    )
        elif parts[0].lower() == "model" and self._model_names:
            partial = parts[1]
            for name in self._model_names:
                if name.startswith(partial):
                    yield Completion(
                        name, start_position=-len(partial), display=_fmt(name)
                    )


class InputSession:
    """Inline input widget: border / input / border / status — no full-screen takeover."""

    _BORDER_STYLE = "fg:#50555a"
    _PROMPT_FT = FormattedText([("fg:#50555a", "  "), ("fg:#787d82", "> ")])

    def __init__(self, completer: Completer) -> None:
        self._completer = completer
        self._history: InMemoryHistory = InMemoryHistory()

    async def prompt_async(self, get_status) -> str:  # get_status: () -> FormattedText
        submitted: list[str | None] = [None]
        show_chrome: list[bool] = [True]

        buf = Buffer(
            completer=self._completer,
            complete_while_typing=True,
            history=self._history,
        )

        kb = KeyBindings()

        @kb.add("enter", eager=True)
        def _enter(event) -> None:  # type: ignore[misc]
            b = event.current_buffer
            if b.complete_state:
                b.complete_state = None
            else:
                submitted[0] = b.text
                b.append_to_history()
                show_chrome[0] = False
                event.app.exit()

        @kb.add("tab", eager=True)
        def _tab(event) -> None:  # type: ignore[misc]
            b = event.current_buffer
            if b.complete_state:
                b.complete_next()
            else:
                b.start_completion(select_first=False)

        @kb.add("s-tab", eager=True)
        def _s_tab(event) -> None:  # type: ignore[misc]
            b = event.current_buffer
            if b.complete_state:
                b.complete_previous()

        @kb.add("c-c", eager=True)
        def _interrupt(event) -> None:  # type: ignore[misc]
            event.app.exit(exception=KeyboardInterrupt())

        @kb.add("c-d", eager=True)
        def _eof(event) -> None:  # type: ignore[misc]
            if not event.current_buffer.text:
                event.app.exit(exception=EOFError())
            else:
                event.current_buffer.delete()

        def _border() -> FormattedText:
            width = shutil.get_terminal_size().columns
            return FormattedText([(InputSession._BORDER_STYLE, "─" * width)])

        input_win = Window(
            BufferControl(buffer=buf, focusable=True),
            get_line_prefix=lambda lineno, wrap_count: (
                InputSession._PROMPT_FT
                if lineno == 0 and wrap_count == 0
                else FormattedText([("", "    ")])
            ),
            wrap_lines=True,
            dont_extend_height=True,
        )

        chrome = Condition(lambda: show_chrome[0])

        layout = Layout(
            HSplit(
                [
                    ConditionalContainer(
                        Window(
                            FormattedTextControl(_border),
                            height=1,
                            dont_extend_height=True,
                        ),
                        filter=chrome,
                    ),
                    input_win,
                    ConditionalContainer(
                        Window(
                            FormattedTextControl(_border),
                            height=1,
                            dont_extend_height=True,
                        ),
                        filter=chrome,
                    ),
                    ConditionalContainer(
                        Window(
                            CompletionsMenuControl(),
                            height=Dimension(min=1, max=5),
                            width=Dimension(weight=1),
                            dont_extend_height=True,
                        ),
                        filter=has_completions & chrome,
                    ),
                    ConditionalContainer(
                        Window(
                            FormattedTextControl(get_status),
                            height=1,
                            dont_extend_height=True,
                        ),
                        filter=~has_completions & chrome,
                    ),
                ]
            ),
            focused_element=input_win,
        )

        app: Application[None] = Application(
            layout=layout,
            key_bindings=merge_key_bindings([load_emacs_bindings(), kb]),
            style=STYLE,
            full_screen=False,
        )

        await app.run_async()

        if submitted[0] is None:
            raise EOFError
        return submitted[0]



# standard
# third party
# custom
from ...conftest import ok, err
from sunwaee.cli import app

WS = "personal"
_BASE = ["--workspace", WS]


def test_create_project(runner, ws):
    data = ok(runner.invoke(app, ["projects", "create", "myproject", *_BASE]))
    assert data["name"] == "myproject"
    assert data["workspace"] == WS
    assert "myproject" in data["path"]
    assert "id" in data
    assert len(data["id"]) == 36


def test_create_project_creates_directory(runner, ws, tmp_path):
    runner.invoke(app, ["projects", "create", "myproject", *_BASE])
    assert (tmp_path / "workspaces" / WS / "projects" / "myproject").is_dir()


def test_create_project_writes_metadata(runner, ws, tmp_path):
    runner.invoke(app, ["projects", "create", "myproject", *_BASE])
    assert (tmp_path / "workspaces" / WS / "projects" / "myproject" / "metadata.md").exists()


def test_create_project_with_body(runner, ws):
    data = ok(runner.invoke(app, ["projects", "create", "myproject", "--body", "My project", *_BASE]))
    assert data["body"] == "My project"


def test_create_duplicate_project_errors(runner, ws):
    runner.invoke(app, ["projects", "create", "myproject", *_BASE])
    e = err(runner.invoke(app, ["projects", "create", "myproject", *_BASE]))
    assert e["code"] == "ALREADY_EXISTS"


def test_create_project_in_specific_workspace(runner, env):
    runner.invoke(app, ["workspaces", "create", "pro"])
    data = ok(runner.invoke(app, ["projects", "create", "alpha", "--workspace", "pro"]))
    assert data["workspace"] == "pro"
    assert data["name"] == "alpha"



# standard
# third party
# custom
from ...conftest import ok, err
from sunwaee.cli import app

WS = "personal"
_BASE = ["--workspace", WS]


def _create_note_and_get_log_id(runner, ws) -> str:
    """Create a note and return the log entry id for that creation."""
    runner.invoke(
        app,
        ["notes", "create", "--title", "Read target", "--body", "x", "--workspace", WS],
    )
    data = ok(runner.invoke(app, ["logs", "list", "--module", "notes", "--operation", "create", *_BASE]))
    entry = next(e for e in data if e.get("item_title") == "Read target")
    return entry["id"]


### READ COMMAND


def test_log_read_by_full_id(runner, ws):
    log_id = _create_note_and_get_log_id(runner, ws)
    data = ok(runner.invoke(app, ["logs", "read", log_id, *_BASE]))
    assert data["id"] == log_id
    assert data["module"] == "notes"
    assert data["operation"] == "create"


def test_log_read_by_id_prefix(runner, ws):
    log_id = _create_note_and_get_log_id(runner, ws)
    prefix = log_id[:8]
    data = ok(runner.invoke(app, ["logs", "read", prefix, *_BASE]))
    assert data["id"] == log_id


def test_log_read_entry_has_all_fields(runner, ws):
    log_id = _create_note_and_get_log_id(runner, ws)
    data = ok(runner.invoke(app, ["logs", "read", log_id, *_BASE]))
    assert "id" in data
    assert "timestamp" in data
    assert "caller" in data
    assert "workspace" in data
    assert "module" in data
    assert "operation" in data
    assert "ok" in data


def test_log_read_not_found(runner, ws):
    e = err(runner.invoke(app, ["logs", "read", "zzz-no-such-id", *_BASE]))
    assert e["code"] == "NOT_FOUND"


def test_log_read_all_workspaces(runner, ws):
    """Read without --workspace searches across all workspaces."""
    log_id = _create_note_and_get_log_id(runner, ws)
    data = ok(runner.invoke(app, ["logs", "read", log_id]))
    assert data["id"] == log_id


def test_log_read_returns_item_title(runner, ws):
    log_id = _create_note_and_get_log_id(runner, ws)
    data = ok(runner.invoke(app, ["logs", "read", log_id, *_BASE]))
    assert data["item_title"] == "Read target"


def test_log_read_entry_id_is_uuid(runner, ws):
    log_id = _create_note_and_get_log_id(runner, ws)
    # UUIDs are 36 chars: 8-4-4-4-12
    assert len(log_id) == 36
    assert log_id.count("-") == 4



# standard
# third party
# custom
from ...conftest import ok, err
from sunwaee.cli import app


def test_read_workspace(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    data = ok(runner.invoke(app, ["workspaces", "read", "personal"]))
    assert data["name"] == "personal"
    assert "id" in data
    assert "body" in data
    assert "path" in data
    assert "default" in data


def test_read_workspace_has_uuid(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    data = ok(runner.invoke(app, ["workspaces", "read", "personal"]))
    assert len(data["id"]) == 36


def test_read_workspace_default_flag(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal"])
    runner.invoke(app, ["workspaces", "create", "pro"])
    runner.invoke(app, ["workspaces", "update", "personal", "--set-default"])
    data = ok(runner.invoke(app, ["workspaces", "read", "personal"]))
    assert data["default"] is True
    data2 = ok(runner.invoke(app, ["workspaces", "read", "pro"]))
    assert data2["default"] is False


def test_read_workspace_with_body(runner, env):
    runner.invoke(app, ["workspaces", "create", "personal", "--body", "My workspace"])
    data = ok(runner.invoke(app, ["workspaces", "read", "personal"]))
    assert data["body"] == "My workspace"


def test_read_nonexistent_workspace_errors(runner, env):
    e = err(runner.invoke(app, ["workspaces", "read", "ghost"]))
    assert e["code"] == "NOT_FOUND"



# standard
from typing import Optional

# third party
import typer

# custom
from sunwaee.config import (
    get_active_project,
    get_default_workspace,
    get_project_dir,
    set_default_project,
)
from sunwaee.core.audit import append_log
from sunwaee.core.fs import read_md, write_md
from sunwaee.core.output import console, error, success
from sunwaee.modules.projects.model import Project


def update(
    name: str = typer.Argument(..., help="Project name"),
    workspace: Optional[str] = typer.Option(
        None, "--workspace", "-w", help="Workspace name (default: from config)"
    ),
    body: Optional[str] = typer.Option(
        None, "--body", "-b", help="Project description / details (markdown)"
    ),
    set_default: bool = typer.Option(False, "--set-default", help="Set as default project"),
) -> None:
    """Update a project."""
    if body is None and not set_default:
        error("Specify at least one of --body or --set-default", "VALIDATION_ERROR")

    ws = workspace or get_default_workspace()
    project_dir = get_project_dir(ws, name)
    if not project_dir.exists():
        error(f"Project not found: {name!r}", "NOT_FOUND")

    meta_path = project_dir / "metadata.md"
    if meta_path.exists():
        meta, existing_body = read_md(meta_path)
        proj = Project.from_dict(name, meta, existing_body)
    else:
        proj = Project(name=name)

    if body is not None:
        proj.body = body
    if set_default:
        set_default_project(ws, name)

    proj.touch()
    write_md(meta_path, proj.to_frontmatter_meta(), proj.body)
    append_log(ws, "projects", "update", item_id=proj.id, item_title=name)

    data = proj.to_dict()
    data["workspace"] = ws
    data["default"] = get_active_project(ws) == name
    success(
        data,
        human_fn=lambda: console.print(
            f"[green]Updated[/green] project '[bold]{name}[/bold]'"
        ),
    )



from __future__ import annotations

import os
import re
import subprocess
from pathlib import Path

from rich.console import Console
from rich.text import Text

from sunwaee.core.logger import get_logger

_log = get_logger("core.git")
_console = Console()

_DIFF_MAX_LINES = 120
_DIFF_EXTENDED_PREFIXES = (
    "diff --git ",
    "index ",
    "new file mode ",
    "deleted file mode ",
    "old mode ",
    "new mode ",
    "similarity index ",
    "rename from ",
    "rename to ",
    "Binary files ",
)
_HUNK_RE = re.compile(r"^@@ -(\d+)(?:,\d+)? \+(\d+)(?:,\d+)? @@")

_GITIGNORE = "logs/\ngen/sessions/\n"
_GIT_ENV = {
    "GIT_AUTHOR_NAME": "SUNWÆE",
    "GIT_AUTHOR_EMAIL": "sunwaee@local",
    "GIT_COMMITTER_NAME": "SUNWÆE",
    "GIT_COMMITTER_EMAIL": "sunwaee@local",
}


def _run(args: list[str], cwd: Path) -> subprocess.CompletedProcess:
    return subprocess.run(
        args,
        cwd=cwd,
        capture_output=True,
        text=True,
        env={**os.environ, **_GIT_ENV},
    )


def ensure_git_repo(ws_dir: Path) -> None:
    """Git-init the workspace directory if it isn't already a repo. Updates .gitignore otherwise."""
    if (ws_dir / ".git").exists():
        _ensure_gitignore_entries(ws_dir)
        return

    _log.info("Initializing git repo at %s", ws_dir)
    ws_dir.mkdir(parents=True, exist_ok=True)
    _run(["git", "init"], cwd=ws_dir)

    gitignore = ws_dir / ".gitignore"
    if not gitignore.exists():
        gitignore.write_text(_GITIGNORE)

    _run(["git", "add", ".gitignore"], cwd=ws_dir)
    _run(["git", "commit", "-m", "init: workspace"], cwd=ws_dir)


def _ensure_gitignore_entries(ws_dir: Path) -> None:
    """Add any missing required entries to an existing .gitignore."""
    gitignore = ws_dir / ".gitignore"
    current = gitignore.read_text() if gitignore.exists() else ""
    required = [line for line in _GITIGNORE.splitlines() if line]
    missing = [e for e in required if e not in current.splitlines()]
    if not missing:
        return
    gitignore.write_text(current.rstrip("\n") + "\n" + "\n".join(missing) + "\n")
    _run(["git", "add", ".gitignore"], cwd=ws_dir)
    _run(["git", "commit", "-m", "chore: update .gitignore"], cwd=ws_dir)


def ws_commit(ws_dir: Path, message: str) -> str:
    """Stage all changes, capture diff, commit. Returns the unified diff string."""
    _run(["git", "add", "-A"], cwd=ws_dir)
    diff = _run(["git", "diff", "--cached"], cwd=ws_dir).stdout
    result = _run(["git", "commit", "-m", message], cwd=ws_dir)
    if result.returncode != 0:
        _log.warning("git commit failed: %s", result.stderr.strip())
    return diff


def render_diff(diff: str) -> None:
    """Render a unified git diff to the terminal with syntax highlighting."""
    all_lines = diff.splitlines()

    old_line = new_line = shown = 0
    break_idx = len(all_lines)

    for i, raw in enumerate(all_lines):
        if shown >= _DIFF_MAX_LINES:
            break_idx = i
            break

        if raw.startswith(_DIFF_EXTENDED_PREFIXES):
            continue

        if raw.startswith("--- "):
            continue

        if raw.startswith("+++ "):
            filename = re.sub(r"^b/", "", raw[4:])
            _console.print()
            _console.print(f"  [bold]{filename}[/bold]")
            continue

        if raw.startswith("@@"):
            m = _HUNK_RE.match(raw)
            if m:
                old_line = int(m.group(1))
                new_line = int(m.group(2))
            _console.print("  [dim]     ...[/dim]")
            shown += 1
            continue

        if raw.startswith("+"):
            t = Text(f"  {new_line:>4} +", style="on #1a3a1a")
            t.append(raw[1:], style="on #1a3a1a")
            t.append(" " * max(0, _console.width - len(t.plain)), style="on #1a3a1a")
            _console.print(t)
            new_line += 1
            shown += 1
            continue

        if raw.startswith("-"):
            t = Text(f"  {old_line:>4} -", style="on #3a1a1a")
            t.append(raw[1:], style="on #3a1a1a")
            t.append(" " * max(0, _console.width - len(t.plain)), style="on #3a1a1a")
            _console.print(t)
            old_line += 1
            shown += 1
            continue

        # context line
        t = Text(f"  {new_line:>4}  ", style="dim")
        t.append(raw[1:])
        _console.print(t)
        old_line += 1
        new_line += 1
        shown += 1

    remaining = sum(
        1
        for l in all_lines[break_idx:]
        if not l.startswith(_DIFF_EXTENDED_PREFIXES) and not l.startswith(("--- ", "+++ "))
    )
    if remaining > 0:
        _console.print(f"  [dim]… {remaining} more lines[/dim]")
    _console.print()


_DEMO_DIFF = """\
diff --git a/sunwaee/core/example.py b/sunwaee/core/example.py
index 1234567..89abcde 100644
--- a/sunwaee/core/example.py
+++ b/sunwaee/core/example.py
@@ -1,10 +1,12 @@
 from __future__ import annotations

-import os
+import os
+import sys
 from pathlib import Path


-def greet(name: str) -> None:
-    print(f"Hello, {name}")
+def greet(name: str, loud: bool = False) -> None:
+    msg = f"Hello, {name}"
+    print(msg.upper() if loud else msg)


 def farewell(name: str) -> None:
diff --git a/sunwaee/core/new_file.py b/sunwaee/core/new_file.py
new file mode 100644
index 0000000..aaaaaaa
--- /dev/null
+++ b/sunwaee/core/new_file.py
@@ -0,0 +1,4 @@
+from __future__ import annotations
+
+
+CONSTANT = 42
"""


if __name__ == "__main__":
    render_diff(_DEMO_DIFF)



# standard
# third party
# custom
from ...conftest import ok, err
from sunwaee.cli import app

WS = "personal"
_BASE = ["--workspace", WS]


def test_read_project(runner, ws):
    runner.invoke(app, ["projects", "create", "myproject", *_BASE])
    data = ok(runner.invoke(app, ["projects", "read", "myproject", *_BASE]))
    assert data["name"] == "myproject"
    assert data["workspace"] == WS
    assert "id" in data
    assert "body" in data
    assert "default" in data


def test_read_project_has_uuid(runner, ws):
    runner.invoke(app, ["projects", "create", "myproject", *_BASE])
    data = ok(runner.invoke(app, ["projects", "read", "myproject", *_BASE]))
    assert len(data["id"]) == 36


def test_read_project_default_flag(runner, ws):
    runner.invoke(app, ["projects", "create", "alpha", *_BASE])
    runner.invoke(app, ["projects", "create", "beta", *_BASE])
    runner.invoke(app, ["projects", "update", "alpha", "--set-default", *_BASE])
    data = ok(runner.invoke(app, ["projects", "read", "alpha", *_BASE]))
    assert data["default"] is True
    data2 = ok(runner.invoke(app, ["projects", "read", "beta", *_BASE]))
    assert data2["default"] is False


def test_read_project_with_body(runner, ws):
    runner.invoke(app, ["projects", "create", "myproject", "--body", "Details", *_BASE])
    data = ok(runner.invoke(app, ["projects", "read", "myproject", *_BASE]))
    assert data["body"] == "Details"


def test_read_nonexistent_project_errors(runner, ws):
    e = err(runner.invoke(app, ["projects", "read", "ghost", *_BASE]))
    assert e["code"] == "NOT_FOUND"



from __future__ import annotations

# standard
import json
import time
from collections.abc import AsyncIterator

# third party
import httpx

# custom
from ..base import BaseEngine
from ..types import (
    Message,
    Performance,
    Response,
    StopReason,
    Tool,
    ToolCall,
    Usage,
    make_non_stream_performance,
    resolve_tokens,
)


class OpenAIEngine(BaseEngine):
    def __init__(
        self,
        model: str,
        api_key: str,
        provider: str = "openai",
        base_url: str = "https://api.openai.com/v1",
        max_tokens: int = 8192,
        reasoning_effort: str | None = None,
    ) -> None:
        self.provider = provider
        self.model = model
        self.api_key = api_key
        self.base_url = base_url.rstrip("/")
        self.max_tokens = max_tokens
        # NOTE "low" | "medium" | "high" — only sent for reasoning-capable models
        self.reasoning_effort = reasoning_effort

    ### PAYLOAD BUILDERS

    def _build_messages(self, messages: list[Message]) -> list[dict]:
        # NOTE reasoning is only echoed on the last assistant turn so that cross-engine
        # conversation history doesn't send provider-specific data to the wrong API.
        last_assistant_idx = max(
            (i for i, m in enumerate(messages) if m.role.value == "assistant"),
            default=-1,
        )

        result = []
        for idx, m in enumerate(messages):
            if m.role.value == "tool":
                result.append(
                    {
                        "role": "tool",
                        "tool_call_id": m.tool_call_id,
                        "content": m.content,
                    }
                )
            elif m.tool_calls:
                msg: dict = {
                    "role": "assistant",
                    "content": m.content,
                    "tool_calls": [
                        {
                            "id": tc.id,
                            "type": "function",
                            "function": {
                                "name": tc.name,
                                "arguments": json.dumps(tc.arguments),
                            },
                        }
                        for tc in m.tool_calls
                    ],
                }
                if idx == last_assistant_idx and m.reasoning_content is not None:
                    msg["reasoning_content"] = m.reasoning_content
                result.append(msg)
            else:
                entry: dict = {"role": m.role.value, "content": m.content}
                if (
                    m.role.value == "assistant"
                    and idx == last_assistant_idx
                    and m.reasoning_content is not None
                ):
                    entry["reasoning_content"] = m.reasoning_content
                result.append(entry)
        return result

    def _build_tools(self, tools: list[Tool]) -> list[dict]:
        return [
            {
                "type": "function",
                "function": {
                    "name": t.name,
                    "description": t.description,
                    "parameters": t.parameters,
                },
            }
            for t in tools
        ]

    def _build_payload(self, messages: list[Message], tools: list[Tool] | None) -> dict:
        payload: dict = {
            "model": self.model,
            "messages": self._build_messages(messages),
            "max_completion_tokens": self.max_tokens,
        }
        if self.reasoning_effort is not None:
            payload["reasoning_effort"] = self.reasoning_effort
        if tools:
            payload["tools"] = self._build_tools(tools)
        return payload

    ### RESPONSE PARSERS

    def _parse_response(self, raw: dict) -> Response:
        choice = raw["choices"][0]
        msg = choice["message"]
        content: str | None = msg.get("content")
        # NOTE deepseek reasoner and other reasoning models expose chain-of-thought here.
        # NOTE Not sent back in subsequent requests — informational only.
        reasoning_content: str | None = msg.get("reasoning_content")
        tool_calls: list[ToolCall] | None = None

        if msg.get("tool_calls"):
            tool_calls = [
                ToolCall(
                    id=tc["id"],
                    name=tc["function"]["name"],
                    arguments=json.loads(tc["function"]["arguments"]),
                )
                for tc in msg["tool_calls"]
            ]

        usage = raw.get("usage", {})
        in_tok = usage.get("prompt_tokens", 0)
        out_tok = usage.get("completion_tokens", 0)
        tot_tok = usage.get("total_tokens", 0)
        cache_read_tokens = usage.get("prompt_tokens_details", {}).get("cached_tokens", 0)
        input_tokens, output_tokens, total_tokens = resolve_tokens(
            in_tok, out_tok, tot_tok
        )
        finish_reason = choice.get("finish_reason", "stop")
        stop_map = {
            "stop": StopReason.END_TURN,
            "tool_calls": StopReason.TOOL_USE,
            "length": StopReason.MAX_TOKENS,
        }

        return Response(
            provider=self.provider,
            model=self.model,
            content=content,
            tool_calls=tool_calls,
            stop_reason=stop_map.get(finish_reason, StopReason.END_TURN),
            reasoning_content=reasoning_content,
            usage=Usage(
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                total_tokens=total_tokens,
                cache_read_tokens=cache_read_tokens,
            ),
        )

    ### INTERFACES

    def _headers(self) -> dict:
        return {"Authorization": f"Bearer {self.api_key}"}

    async def chat(
        self,
        messages: list[Message],
        tools: list[Tool] | None = None,
    ) -> Response:
        t_start = time.monotonic()
        async with httpx.AsyncClient() as client:
            resp = await client.post(
                f"{self.base_url}/chat/completions",
                headers=self._headers(),
                json=self._build_payload(messages, tools),
                timeout=120,
            )
            if not resp.is_success:
                raise RuntimeError(
                    f"OpenAI-compatible API error {resp.status_code}: {resp.text}"
                )
            response = self._parse_response(resp.json())
        total_duration = time.monotonic() - t_start
        response.performance = make_non_stream_performance(response, total_duration)
        return response

    async def stream(
        self,
        messages: list[Message],
        tools: list[Tool] | None = None,
    ) -> AsyncIterator[Response]:
        payload = self._build_payload(messages, tools)
        payload["stream"] = True
        payload["stream_options"] = {"include_usage": True}

        stop_reason: str = StopReason.END_TURN.value
        stop_map = {
            "stop": StopReason.END_TURN,
            "tool_calls": StopReason.TOOL_USE,
            "length": StopReason.MAX_TOKENS,
        }
        input_tokens = 0
        output_tokens = 0
        total_tokens = 0
        cache_read_tokens = 0

        # tool call accumulators keyed by index inside choices[0].delta.tool_calls.
        tool_buf: dict[int, dict] = {}

        t_start = time.monotonic()
        t_first_chunk: float | None = None
        t_reasoning_start: float | None = None
        t_reasoning_end: float | None = None
        t_content_start: float | None = None

        async with httpx.AsyncClient() as client:
            async with client.stream(
                "POST",
                f"{self.base_url}/chat/completions",
                headers=self._headers(),
                json=payload,
                timeout=120,
            ) as resp:
                resp.raise_for_status()
                if self.reasoning_effort is not None:
                    # Model will reason silently — tokens are not streamed.
                    # Yield a sentinel immediately so the consumer isn't left in
                    # the dark; record the request time as reasoning start.
                    now = time.monotonic()
                    t_first_chunk = now
                    t_reasoning_start = t_start
                    yield Response(
                        provider=self.provider,
                        model=self.model,
                        streaming=True,
                        synthetic=True,
                        reasoning_content="Reasoning in progress - tokens not available for this model",
                    )
                async for line in resp.aiter_lines():
                    if not line.startswith("data: "):  # pragma: no cover
                        continue
                    data = line[6:]
                    if data == "[DONE]":
                        break
                    now = time.monotonic()
                    if t_first_chunk is None:
                        t_first_chunk = now

                    chunk = json.loads(data)
                    # With include_usage, a final chunk arrives with usage and no choices.
                    if usage := chunk.get("usage"):
                        in_tok = usage.get("prompt_tokens", 0)
                        out_tok = usage.get("completion_tokens", 0)
                        tot_tok = usage.get("total_tokens", 0)
                        cache_read_tokens = usage.get("prompt_tokens_details", {}).get("cached_tokens", 0)
                        input_tokens, output_tokens, total_tokens = resolve_tokens(
                            in_tok, out_tok, tot_tok
                        )
                    if not chunk.get("choices"):
                        continue
                    choice = chunk["choices"][0]
                    delta = choice.get("delta", {})

                    if fr := choice.get("finish_reason"):
                        if tool_buf:
                            tcs = [
                                ToolCall(
                                    id=buf.get("id") or f"tc_{i}",
                                    name=buf["name"],
                                    arguments=(
                                        json.loads(buf["arguments"])
                                        if buf["arguments"]
                                        else {}
                                    ),
                                )
                                for i, buf in sorted(tool_buf.items())
                            ]
                            stop_reason = fr
                            yield Response(
                                provider=self.provider,
                                model=self.model,
                                streaming=True,
                                tool_calls=tcs,
                            )
                            tool_buf.clear()

                    if rc := delta.get("reasoning_content"):
                        if t_reasoning_start is None:
                            t_reasoning_start = now
                        yield Response(
                            provider=self.provider,
                            model=self.model,
                            streaming=True,
                            reasoning_content=rc,
                        )

                    if text := delta.get("content"):
                        if t_content_start is None:
                            t_content_start = now
                            if (
                                t_reasoning_start is not None
                                and t_reasoning_end is None
                            ):
                                t_reasoning_end = now
                        yield Response(
                            provider=self.provider,
                            model=self.model,
                            streaming=True,
                            content=text,
                        )

                    for tc_delta in delta.get("tool_calls", []):
                        i = tc_delta.get("index", 0)
                        if i not in tool_buf:
                            tool_buf[i] = {"id": "", "name": "", "arguments": ""}
                        if tc_id := tc_delta.get("id"):
                            tool_buf[i]["id"] = tc_id
                        fn = tc_delta.get("function", {})
                        if name := fn.get("name"):
                            tool_buf[i]["name"] = name
                        if args := fn.get("arguments"):
                            tool_buf[i]["arguments"] += args

        t_end = time.monotonic()
        total_duration = t_end - t_start
        latency = (t_first_chunk - t_start) if t_first_chunk else total_duration
        reasoning_duration = (
            ((t_reasoning_end or t_end) - t_reasoning_start)
            if t_reasoning_start
            else 0.0
        )
        content_duration = (t_end - t_content_start) if t_content_start else 0.0
        throughput = int(output_tokens / total_duration) if total_duration > 0 else 0

        yield Response(
            provider=self.provider,
            model=self.model,
            streaming=True,
            stop_reason=stop_map.get(stop_reason, StopReason.END_TURN),
            usage=Usage(
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                total_tokens=total_tokens,
                cache_read_tokens=cache_read_tokens,
            ),
            performance=Performance(
                latency=round(latency, 2),
                reasoning_duration=round(reasoning_duration, 2),
                content_duration=round(content_duration, 2),
                total_duration=round(total_duration, 2),
                throughput=throughput,
            ),
        )



# standard
import os
import tomllib
import tomli_w
from enum import Enum
from pathlib import Path

# third party
# custom
from sunwaee.core.logger import get_logger

_log = get_logger("config")

_DEFAULT_CONFIG: dict = {
    "workspaces": {
        "default": "personal",
        "base_dir": "~/sunwaee/workspaces",
    },
    "sun": {
        "default_model": "claude-sonnet-4-6",
        "default_provider": "anthropic",
        "api_url": "http://localhost:8000",
    },
}


class Caller(str, Enum):
    USER = "user"
    API = "api"
    SUN = "sun"


def _config_dir() -> Path:
    return Path(os.environ.get("SUNWAEE_CONFIG_DIR", Path.home() / ".sunwaee"))


def _config_file() -> Path:
    return _config_dir() / "config.toml"


def get_caller() -> Caller:
    val = os.environ.get("SUNWAEE_CALLER", "human").lower()
    try:
        return Caller(val)
    except ValueError:
        return Caller.USER


def is_machine_caller() -> bool:
    return get_caller() in (Caller.API, Caller.SUN)


def _load_config() -> dict:
    cfg_dir = _config_dir()
    cfg_file = _config_file()
    if not cfg_file.exists():
        _log.info("Config file not found, creating default at %s", cfg_file)
        cfg_dir.mkdir(parents=True, exist_ok=True)
        with open(cfg_file, "wb") as f:
            tomli_w.dump(_DEFAULT_CONFIG, f)
    _log.debug("Loading config from %s", cfg_file)
    with open(cfg_file, "rb") as f:
        return tomllib.load(f)


def _save_config(cfg: dict) -> None:
    cfg_dir = _config_dir()
    cfg_dir.mkdir(parents=True, exist_ok=True)
    _log.debug("Saving config to %s", _config_file())
    with open(_config_file(), "wb") as f:
        tomli_w.dump(cfg, f)


def get_workspaces_base_dir() -> Path:
    env = os.environ.get("SUNWAEE_WORKSPACES_DIR")
    if env:
        return Path(env).expanduser()
    cfg = _load_config()
    return Path(
        cfg.get("workspaces", {}).get("base_dir", "~/sunwaee/workspaces")
    ).expanduser()


def get_default_workspace() -> str:
    env = os.environ.get("SUNWAEE_WORKSPACE")
    if env:
        return env
    cfg = _load_config()
    return cfg.get("workspaces", {}).get("default", "personal")


def set_default_workspace(name: str) -> None:
    _log.info("Setting default workspace to %r", name)
    cfg = _load_config()
    cfg.setdefault("workspaces", {})["default"] = name
    _save_config(cfg)


def get_sun_default_model() -> tuple[str, str]:
    """Return (model_name, provider) from config."""
    cfg = _load_config()
    sun = cfg.get("sun", {})
    model = sun.get("default_model", "claude-sonnet-4-6")
    provider = sun.get("default_provider", "anthropic")
    return model, provider


def set_sun_default_model(model: str, provider: str) -> None:
    cfg = _load_config()
    cfg.setdefault("sun", {})["default_model"] = model
    cfg.setdefault("sun", {})["default_provider"] = provider
    _save_config(cfg)


def get_all_workspaces() -> list[str]:
    """Return all workspace names found in the base directory."""
    base = get_workspaces_base_dir()
    if not base.exists():
        return [get_default_workspace()]
    dirs = sorted(d.name for d in base.iterdir() if d.is_dir())
    return dirs if dirs else [get_default_workspace()]


def get_workspace_dir(workspace: str) -> Path:
    """Return the root directory for a given workspace."""
    return get_workspaces_base_dir() / workspace


def get_module_dir(workspace: str, module: str, project: str | None = None) -> Path:
    """Return the directory for a given module within a workspace (or project)."""
    base = get_workspace_dir(workspace)
    if project:
        return base / "projects" / project / module
    return base / module


def get_project_dir(workspace: str, project: str) -> Path:
    """Return the root directory for a given project within a workspace."""
    return get_workspace_dir(workspace) / "projects" / project


def get_all_projects(workspace: str) -> list[str]:
    """Return all project names within a workspace."""
    projects_dir = get_workspace_dir(workspace) / "projects"
    if not projects_dir.exists():
        return []
    return sorted(
        d.name
        for d in projects_dir.iterdir()
        if d.is_dir() and not d.name.startswith(".")
    )


def get_active_project(workspace: str | None = None) -> str | None:
    """Return the active project for the given workspace, or None if workspace-level."""
    env = os.environ.get("SUNWAEE_PROJECT")
    if env:
        return env
    ws = workspace or get_default_workspace()
    cfg = _load_config()
    return cfg.get("projects", {}).get(ws)


def set_default_project(workspace: str, name: str) -> None:
    cfg = _load_config()
    cfg.setdefault("projects", {})[workspace] = name
    _save_config(cfg)



# standard
import uuid
from dataclasses import dataclass, field
from typing import Optional

# third party
# custom


def _uuid() -> str:
    return str(uuid.uuid4())


@dataclass
class LogEntry:
    timestamp: str
    caller: str
    workspace: str
    module: str
    operation: str
    ok: bool = True
    item_id: Optional[str] = None
    item_title: Optional[str] = None
    args: Optional[dict] = None
    id: str = field(default_factory=_uuid)

    def to_dict(self) -> dict:
        d: dict = {
            "id": self.id,
            "timestamp": self.timestamp,
            "caller": self.caller,
            "workspace": self.workspace,
            "module": self.module,
            "operation": self.operation,
            "ok": self.ok,
        }
        if self.item_id is not None:
            d["item_id"] = self.item_id
        if self.item_title is not None:
            d["item_title"] = self.item_title
        if self.args:
            d["args"] = self.args
        return d

    def matches(self, query: str) -> bool:
        """Case-insensitive substring match across searchable fields."""
        q = query.lower()
        return any(
            q in (f or "").lower()
            for f in [
                self.id,
                self.item_title,
                self.item_id,
                self.module,
                self.operation,
                self.caller,
                self.workspace,
            ]
        )

    @classmethod
    def from_dict(cls, d: dict) -> "LogEntry":
        return cls(
            timestamp=d.get("timestamp", ""),
            caller=d.get("caller", ""),
            workspace=d.get("workspace", ""),
            module=d.get("module", ""),
            operation=d.get("operation", ""),
            ok=d.get("ok", True),
            item_id=d.get("item_id"),
            item_title=d.get("item_title"),
            args=d.get("args"),
            id=d.get("id") or _uuid(),
        )



import json

from sunwaee.modules.shell.tools import shell, TOOLS as SHELL_TOOLS


def test_shell_run_success(ws):
    data = json.loads(shell(command="echo hello", comment="Print hello"))
    assert data["ok"] is True
    assert data["data"]["exit_code"] == 0
    assert "hello" in data["data"]["stdout"]
    assert data["data"]["comment"] == "Print hello"
    assert data["data"]["command"] == "echo hello"


def test_shell_run_stderr(ws):
    data = json.loads(shell(command="echo error >&2", comment="Write to stderr"))
    assert data["ok"] is True
    assert "error" in data["data"]["stderr"]


def test_shell_run_nonzero_exit(ws):
    data = json.loads(shell(command="exit 1", comment="Exit with error"))
    assert data["ok"] is True
    assert data["data"]["exit_code"] == 1


def test_shell_run_multiline_output(ws):
    data = json.loads(shell(command="printf 'a\\nb\\nc\\n'", comment="Multi-line output"))
    assert data["ok"] is True
    lines = data["data"]["stdout"].strip().splitlines()
    assert lines == ["a", "b", "c"]


def test_shell_run_invalid_command(ws):
    data = json.loads(shell(command="nonexistent_command_xyz", comment="Invalid command"))
    assert data["ok"] is True
    assert data["data"]["exit_code"] != 0
    assert data["data"]["stderr"] != ""


def test_shell_run_git_snapshot_fields(ws):
    data = json.loads(shell(command="echo x", comment="Check git fields"))
    assert data["ok"] is True
    assert "git_diff" in data["data"]
    assert "git_status" in data["data"]


def test_shell_tools_registry():
    assert len(SHELL_TOOLS) == 1
    assert SHELL_TOOLS[0].name == "shell"
    assert SHELL_TOOLS[0].fn is not None
    props = SHELL_TOOLS[0].parameters["properties"]
    assert "command" in props
    assert "comment" in props
